"""Methods to load products in elasticsearch
"""
import re
from typing import Iterable

from robotoff import settings
from robotoff.products import ProductDataset
from robotoff.spellcheck.items import Ingredients
from robotoff.utils import get_logger

logger = get_logger(__name__)


MULTIPLE_SPACES_RE = re.compile(r"\s{2,}")


def generate_product_data() -> Iterable[tuple[str, dict]]:
    dataset = ProductDataset(settings.JSONL_DATASET_PATH)

    product_stream = (
        dataset.stream()
        .filter_text_field("lang", "fr")
        .filter_by_country_tag("en:france")
        .filter_nonempty_text_field("ingredients_text_fr")
        .filter_by_state_tag("en:complete")
    )

    product_iter = product_stream.iter()
    product_iter = (
        p for p in product_iter if int(p.get("unknown_ingredients_n", 0)) == 0
    )

    return (
        (
            product["code"],
            {
                "ingredients_text_fr": normalize_ingredient_list(
                    product["ingredients_text_fr"]
                )
            },
        )
        for product in product_iter
    )


def empty_ingredient(ingredient: str) -> bool:
    return not bool(ingredient.strip(" /-.%0123456789"))


def normalize_ingredient_list(ingredient_text: str) -> list[str]:
    ingredients = Ingredients.from_text(ingredient_text)

    normalized = []

    for ingredient in ingredients.get_iter():
        if empty_ingredient(ingredient):
            continue

        ingredient = MULTIPLE_SPACES_RE.sub(" ", ingredient)
        ingredient = ingredient.strip(" .")
        normalized.append(ingredient)

    return normalized



from typing import Optional, Union

from flashtext import KeywordProcessor

from robotoff import settings
from robotoff.prediction.types import Prediction
from robotoff.types import PredictionType
from robotoff.utils import text_file_iter

from .dataclass import OCRResult, SafeSearchAnnotationLikelihood, get_text

LABELS_TO_FLAG = {
    "Face",
    "Head",
    "Selfie",
    "Hair",
    "Forehead",
    "Chin",
    "Cheek",
    "Arm",
    "Tooth",
    "Human Leg",
    "Ankle",
    "Eyebrow",
    "Ear",
    "Neck",
    "Jaw",
    "Nose",
    "Facial Expression",
    "Glasses",
    "Eyewear",
    # Gesture generate too many false positive
    # "Gesture",
    # Thumb is pretty common on OFF images (as products are often hold in
    # hands)
    # "Thumb",
    "Jeans",
    "Shoe",
    "Child",
    "Baby",
    "Human",
    "Dog",
    "Cat",
    "Computer",
    "Laptop",
    "Refrigerator",
    "Cat",  # https://world.openfoodfacts.org/images/products/761/002/911/3600/1.json
}


def generate_image_flag_keyword_processor() -> KeywordProcessor:
    processor = KeywordProcessor()

    for key, file_path in (
        ("beauty", settings.OCR_IMAGE_FLAG_BEAUTY_PATH),
        ("miscellaneous", settings.OCR_IMAGE_FLAG_MISCELLANEOUS_PATH),
    ):
        for name in text_file_iter(file_path):
            processor.add_keyword(name, clean_name=(name, key))

    return processor


PROCESSOR = generate_image_flag_keyword_processor()


def extract_image_flag_flashtext(
    processor: KeywordProcessor, text: str
) -> Optional[Prediction]:
    for (_, key), span_start, span_end in processor.extract_keywords(
        text, span_info=True
    ):
        match_str = text[span_start:span_end]
        return Prediction(
            type=PredictionType.image_flag,
            data={"text": match_str, "type": "text", "label": key},
        )

    return None


def flag_image(content: Union[OCRResult, str]) -> list[Prediction]:
    predictions: list[Prediction] = []

    text = get_text(content)
    prediction = extract_image_flag_flashtext(PROCESSOR, text)

    if prediction is not None:
        predictions.append(prediction)

    if isinstance(content, str):
        return predictions

    safe_search_annotation = content.get_safe_search_annotation()
    label_annotations = content.get_label_annotations()

    if safe_search_annotation:
        for key in ("adult", "violence"):
            value: SafeSearchAnnotationLikelihood = getattr(safe_search_annotation, key)
            if value >= SafeSearchAnnotationLikelihood.VERY_LIKELY:
                predictions.append(
                    Prediction(
                        type=PredictionType.image_flag,
                        data={
                            "type": "safe_search_annotation",
                            "label": key,
                            "likelihood": value.name,
                        },
                    )
                )

    for label_annotation in label_annotations:
        if (
            label_annotation.description in LABELS_TO_FLAG
            and label_annotation.score >= 0.6
        ):
            predictions.append(
                Prediction(
                    type=PredictionType.image_flag,
                    data={
                        "type": "label_annotation",
                        "label": label_annotation.description.lower(),
                        "likelihood": label_annotation.score,
                    },
                )
            )
            break

    return predictions



from collections import defaultdict
from pathlib import Path
from typing import AbstractSet, Callable, Optional

from robotoff.settings import (
    FR_TOKENS_PATH,
    INGREDIENT_TOKENS_PATH,
    INGREDIENTS_FR_PATH,
)
from robotoff.utils import text_file_iter
from robotoff.utils.cache import CachedStore

DEACCENTED_TOKENS = dict[str, list[str]]


class Vocabulary(object):
    def __init__(
        self,
        voc_name: Optional[str] = None,
        tokens: Optional[AbstractSet] = None,
        deaccented_tokens: Optional[DEACCENTED_TOKENS] = None,
    ):
        self.voc = set()

        if voc_name is not None:
            self.voc = VOC_CACHE[
                voc_name
            ].get()  # CACHE is defined after Vocabulary class

        if tokens is not None:
            self.voc.update(tokens)

        if voc_name is not None:
            assert deaccented_tokens is None
            self.deaccented_tokens = DEACCENTED_TOKENS_CACHE[voc_name].get()
        elif deaccented_tokens is not None:
            self.deaccented_tokens = defaultdict(list, deaccented_tokens)
        else:
            self.deaccented_tokens = self.deaccent_tokens_fn(self.voc)()

    def __contains__(self, token: str) -> bool:
        return self.normalize(token) in self.voc

    def _contains_deaccent(self, token: str) -> bool:
        return self.deaccent(token) in self.deaccented_tokens

    def __or__(self, other):
        return Vocabulary(
            tokens=self.voc | other.voc,
            deaccented_tokens=defaultdict(
                list, dict(self.deaccented_tokens, **other.deaccented_tokens)
            ),
        )

    def suggest(self, token: str) -> Optional[str]:
        deaccent_suggestions: Optional[list[str]] = self._suggest_deaccent(token)
        if deaccent_suggestions is not None:
            if len(deaccent_suggestions) == 1:
                return deaccent_suggestions[0]
            elif len(deaccent_suggestions) > 1:
                return None

        split_suggestions: list[tuple[str, str]] = self._suggest_split(token)
        if len(split_suggestions) == 1:
            return split_suggestions[0][0] + " " + split_suggestions[0][1]
        return None

    def _suggest_deaccent(self, token: str) -> Optional[list[str]]:
        if token in self:
            return None
        deaccented_token = self.deaccent(token)
        if deaccented_token in self.deaccented_tokens:
            return self.deaccented_tokens[deaccented_token]
        return None

    def _suggest_split(self, token: str) -> list[tuple[str, str]]:
        if token in self:
            return []

        suggestions = []
        for i in range(2, len(token) - 1):
            # pre and post must be at least 2 letters long
            pre = token[:i]
            post = token[i:]
            if pre in self and post in self or (pre + " " + post) in self:
                suggestions.append((pre, post))
            else:
                if self._contains_deaccent(pre) and post in self:
                    pre_suggestion = self.deaccented_tokens[self.deaccent(pre)]
                    if len(pre_suggestion) == 1:
                        suggestions.append((pre_suggestion[0], post))

                if pre in self and self._contains_deaccent(post):
                    post_suggestion = self.deaccented_tokens[self.deaccent(post)]
                    if len(post_suggestion) == 1:
                        suggestions.append((pre, post_suggestion[0]))

                if self._contains_deaccent(pre) and self._contains_deaccent(post):
                    pre_suggestion = self.deaccented_tokens[self.deaccent(pre)]
                    post_suggestion = self.deaccented_tokens[self.deaccent(post)]
                    if len(pre_suggestion) == 1 and len(post_suggestion) == 1:
                        suggestions.append((pre_suggestion[0], post_suggestion[0]))
        return suggestions

    @staticmethod
    def deaccent(token: str) -> str:
        ACCENTS = {"a": "à", "e": "éêè", "u": "ùüû", "i": "ïî"}
        for letter in ACCENTS:
            for c in ACCENTS[letter]:
                token = token.replace(c, letter)
        return token

    @staticmethod
    def normalize(token: str) -> str:
        return token.lower()

    @classmethod
    def load_vocabulary_fn(cls, voc_path: Path) -> Callable[[], AbstractSet[str]]:
        def inner_fn() -> AbstractSet[str]:
            return set(cls.normalize(token) for token in text_file_iter(voc_path))

        return inner_fn

    @classmethod
    def deaccent_tokens_fn(cls, voc: AbstractSet) -> Callable[[], DEACCENTED_TOKENS]:
        def inner_fn() -> DEACCENTED_TOKENS:
            deaccented_tokens = defaultdict(list)
            for token in voc:
                deaccented_token = cls.deaccent(token)
                if deaccented_token != token:
                    deaccented_tokens[deaccented_token].append(token)
            return deaccented_tokens

        return inner_fn


def get_voc_cache(path: Path) -> CachedStore:
    return CachedStore(Vocabulary.load_vocabulary_fn(path))


def get_deaccent_cache(voc_cache: CachedStore) -> CachedStore:
    return CachedStore(Vocabulary.deaccent_tokens_fn(voc_cache.get()))


VOC_CACHE: dict[str, CachedStore] = {
    "wikipedia_lower": get_voc_cache(FR_TOKENS_PATH),
    "ingredients_fr": get_voc_cache(INGREDIENTS_FR_PATH),
    "ingredients_fr_tokens": get_voc_cache(INGREDIENT_TOKENS_PATH),
}

DEACCENTED_TOKENS_CACHE: dict[str, CachedStore] = {}
for key, voc_cache in VOC_CACHE.items():
    DEACCENTED_TOKENS_CACHE[key] = get_deaccent_cache(voc_cache)



from robotoff.insights.importer import refresh_insights
from robotoff.models import Prediction, ProductInsight, with_db
from robotoff.products import fetch_dataset, has_dataset_changed
from robotoff.utils import get_logger

from .import_image import run_import_image_job  # noqa: F401
from .product_updated import update_insights_job  # noqa: F401

logger = get_logger(__name__)


@with_db
def download_product_dataset_job():
    """This job is triggered via /api/v1/products/dataset and causes Robotoff
    to re-import the Product Opener product dump."""
    if has_dataset_changed():
        fetch_dataset()


@with_db
def delete_product_insights_job(barcode: str, server_domain: str):
    """This job is triggered by Product Opener via /api/v1/webhook/product
    when the given product has been removed from the database - in this case
    we must delete all of the associated predictions and insights that have
    not been annotated.
    """
    logger.info("Product %s deleted, deleting associated insights...", barcode)
    deleted_predictions = (
        Prediction.delete()
        .where(
            Prediction.barcode == barcode,
            Prediction.server_domain == server_domain,
        )
        .execute()
    )
    deleted_insights = (
        ProductInsight.delete()
        .where(
            ProductInsight.barcode == barcode,
            ProductInsight.annotation.is_null(),
            ProductInsight.server_domain == server_domain,
        )
        .execute()
    )

    logger.info(
        f"{deleted_predictions} predictions deleted, "
        f"{deleted_insights} insights deleted"
    )


@with_db
def refresh_insights_job(barcodes: list[str], server_domain: str):
    logger.info(
        f"Refreshing insights for {len(barcodes)} products, server_domain: {server_domain}"
    )
    for barcode in barcodes:
        refresh_insights(barcode, server_domain)



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""A set of functions that are used for visualization.

These functions often receive an image, perform some visualization on the image.
The functions do not return a value, instead they modify the image itself.

"""
import collections
import warnings

# Set headless-friendly backend.
try:
    import matplotlib
except ImportError:
    warnings.warn("matplotlib library not installed", ImportWarning)
else:
    matplotlib.use("Agg")  # pylint: disable=multiple-statements

import numpy as np
import PIL.Image as Image
import PIL.ImageColor as ImageColor
import PIL.ImageDraw as ImageDraw
import PIL.ImageFont as ImageFont
import six

_TITLE_LEFT_MARGIN = 10
_TITLE_TOP_MARGIN = 10
STANDARD_COLORS = [
    "AliceBlue",
    "Chartreuse",
    "Aqua",
    "Aquamarine",
    "Azure",
    "Beige",
    "Bisque",
    "BlanchedAlmond",
    "BlueViolet",
    "BurlyWood",
    "CadetBlue",
    "AntiqueWhite",
    "Chocolate",
    "Coral",
    "CornflowerBlue",
    "Cornsilk",
    "Crimson",
    "Cyan",
    "DarkCyan",
    "DarkGoldenRod",
    "DarkGrey",
    "DarkKhaki",
    "DarkOrange",
    "DarkOrchid",
    "DarkSalmon",
    "DarkSeaGreen",
    "DarkTurquoise",
    "DarkViolet",
    "DeepPink",
    "DeepSkyBlue",
    "DodgerBlue",
    "FireBrick",
    "FloralWhite",
    "ForestGreen",
    "Fuchsia",
    "Gainsboro",
    "GhostWhite",
    "Gold",
    "GoldenRod",
    "Salmon",
    "Tan",
    "HoneyDew",
    "HotPink",
    "IndianRed",
    "Ivory",
    "Khaki",
    "Lavender",
    "LavenderBlush",
    "LawnGreen",
    "LemonChiffon",
    "LightBlue",
    "LightCoral",
    "LightCyan",
    "LightGoldenRodYellow",
    "LightGray",
    "LightGrey",
    "LightGreen",
    "LightPink",
    "LightSalmon",
    "LightSeaGreen",
    "LightSkyBlue",
    "LightSlateGray",
    "LightSlateGrey",
    "LightSteelBlue",
    "LightYellow",
    "Lime",
    "LimeGreen",
    "Linen",
    "Magenta",
    "MediumAquaMarine",
    "MediumOrchid",
    "MediumPurple",
    "MediumSeaGreen",
    "MediumSlateBlue",
    "MediumSpringGreen",
    "MediumTurquoise",
    "MediumVioletRed",
    "MintCream",
    "MistyRose",
    "Moccasin",
    "NavajoWhite",
    "OldLace",
    "Olive",
    "OliveDrab",
    "Orange",
    "OrangeRed",
    "Orchid",
    "PaleGoldenRod",
    "PaleGreen",
    "PaleTurquoise",
    "PaleVioletRed",
    "PapayaWhip",
    "PeachPuff",
    "Peru",
    "Pink",
    "Plum",
    "PowderBlue",
    "Purple",
    "Red",
    "RosyBrown",
    "RoyalBlue",
    "SaddleBrown",
    "Green",
    "SandyBrown",
    "SeaGreen",
    "SeaShell",
    "Sienna",
    "Silver",
    "SkyBlue",
    "SlateBlue",
    "SlateGray",
    "SlateGrey",
    "Snow",
    "SpringGreen",
    "SteelBlue",
    "GreenYellow",
    "Teal",
    "Thistle",
    "Tomato",
    "Turquoise",
    "Violet",
    "Wheat",
    "White",
    "WhiteSmoke",
    "Yellow",
    "YellowGreen",
]


def save_image_array_as_png(image, output_path):
    """Saves an image (represented as a numpy array) to PNG.

    Args:
      image: a numpy array with shape [height, width, 3].
      output_path: path to which image should be written.
    """
    image_pil = Image.fromarray(np.uint8(image)).convert("RGB")
    with open(output_path, "w") as fid:
        image_pil.save(fid, "PNG")


def encode_image_array_as_png_str(image):
    """Encodes a numpy array into a PNG string.

    Args:
      image: a numpy array with shape [height, width, 3].

    Returns:
      PNG encoded image string.
    """
    image_pil = Image.fromarray(np.uint8(image))
    output = six.BytesIO()
    image_pil.save(output, format="PNG")
    png_string = output.getvalue()
    output.close()
    return png_string


def draw_bounding_box_on_image_array(
    image,
    ymin,
    xmin,
    ymax,
    xmax,
    color="red",
    thickness=4,
    display_str_list=(),
    use_normalized_coordinates=True,
):
    """Adds a bounding box to an image (numpy array).

    Bounding box coordinates can be specified in either absolute (pixel) or
    normalized coordinates by setting the use_normalized_coordinates argument.

    Args:
      image: a numpy array with shape [height, width, 3].
      ymin: ymin of bounding box.
      xmin: xmin of bounding box.
      ymax: ymax of bounding box.
      xmax: xmax of bounding box.
      color: color to draw bounding box. Default is red.
      thickness: line thickness. Default value is 4.
      display_str_list: list of strings to display in box
                        (each to be shown on its own line).
      use_normalized_coordinates: If True (default), treat coordinates
        ymin, xmin, ymax, xmax as relative to the image.  Otherwise treat
        coordinates as absolute.
    """
    image_pil = Image.fromarray(np.uint8(image)).convert("RGB")
    draw_bounding_box_on_image(
        image_pil,
        ymin,
        xmin,
        ymax,
        xmax,
        color,
        thickness,
        display_str_list,
        use_normalized_coordinates,
    )
    np.copyto(image, np.array(image_pil))


def draw_bounding_box_on_image(
    image,
    ymin,
    xmin,
    ymax,
    xmax,
    color="red",
    thickness=4,
    display_str_list=(),
    use_normalized_coordinates=True,
):
    """Adds a bounding box to an image.

    Bounding box coordinates can be specified in either absolute (pixel) or
    normalized coordinates by setting the use_normalized_coordinates argument.

    Each string in display_str_list is displayed on a separate line above the
    bounding box in black text on a rectangle filled with the input 'color'.
    If the top of the bounding box extends to the edge of the image, the strings
    are displayed below the bounding box.

    Args:
      image: a PIL.Image object.
      ymin: ymin of bounding box.
      xmin: xmin of bounding box.
      ymax: ymax of bounding box.
      xmax: xmax of bounding box.
      color: color to draw bounding box. Default is red.
      thickness: line thickness. Default value is 4.
      display_str_list: list of strings to display in box
                        (each to be shown on its own line).
      use_normalized_coordinates: If True (default), treat coordinates
        ymin, xmin, ymax, xmax as relative to the image.  Otherwise treat
        coordinates as absolute.
    """
    draw = ImageDraw.Draw(image)
    im_width, im_height = image.size
    if use_normalized_coordinates:
        (left, right, top, bottom) = (
            xmin * im_width,
            xmax * im_width,
            ymin * im_height,
            ymax * im_height,
        )
    else:
        (left, right, top, bottom) = (xmin, xmax, ymin, ymax)
    draw.line(
        [(left, top), (left, bottom), (right, bottom), (right, top), (left, top)],
        width=thickness,
        fill=color,
    )
    try:
        font = ImageFont.truetype("arial.ttf", 24)
    except IOError:
        font = ImageFont.load_default()

    # If the total height of the display strings added to the top of the bounding
    # box exceeds the top of the image, stack the strings below the bounding box
    # instead of above.
    display_str_heights = [font.getsize(ds)[1] for ds in display_str_list]
    # Each display_str has a top and bottom margin of 0.05x.
    total_display_str_height = (1 + 2 * 0.05) * sum(display_str_heights)

    if top > total_display_str_height:
        text_bottom = top
    else:
        text_bottom = bottom + total_display_str_height
    # Reverse list and print from bottom to top.
    for display_str in display_str_list[::-1]:
        text_width, text_height = font.getsize(display_str)
        margin = np.ceil(0.05 * text_height)
        draw.rectangle(
            [
                (left, text_bottom - text_height - 2 * margin),
                (left + text_width, text_bottom),
            ],
            fill=color,
        )
        draw.text(
            (left + margin, text_bottom - text_height - margin),
            display_str,
            fill="black",
            font=font,
        )
        text_bottom -= text_height - 2 * margin


def draw_bounding_boxes_on_image_array(
    image, boxes, color="red", thickness=4, display_str_list_list=()
):
    """Draws bounding boxes on image (numpy array).

    Args:
      image: a numpy array object.
      boxes: a 2 dimensional numpy array of [N, 4]: (ymin, xmin, ymax, xmax).
             The coordinates are in normalized format between [0, 1].
      color: color to draw bounding box. Default is red.
      thickness: line thickness. Default value is 4.
      display_str_list_list: list of list of strings.
                             a list of strings for each bounding box.
                             The reason to pass a list of strings for a
                             bounding box is that it might contain
                             multiple labels.

    Raises:
      ValueError: if boxes is not a [N, 4] array
    """
    image_pil = Image.fromarray(image)
    draw_bounding_boxes_on_image(
        image_pil, boxes, color, thickness, display_str_list_list
    )
    np.copyto(image, np.array(image_pil))


def draw_bounding_boxes_on_image(
    image, boxes, color="red", thickness=4, display_str_list_list=()
):
    """Draws bounding boxes on image.

    Args:
      image: a PIL.Image object.
      boxes: a 2 dimensional numpy array of [N, 4]: (ymin, xmin, ymax, xmax).
             The coordinates are in normalized format between [0, 1].
      color: color to draw bounding box. Default is red.
      thickness: line thickness. Default value is 4.
      display_str_list_list: list of list of strings.
                             a list of strings for each bounding box.
                             The reason to pass a list of strings for a
                             bounding box is that it might contain
                             multiple labels.

    Raises:
      ValueError: if boxes is not a [N, 4] array
    """
    boxes_shape = boxes.shape
    if not boxes_shape:
        return
    if len(boxes_shape) != 2 or boxes_shape[1] != 4:
        raise ValueError("Input must be of size [N, 4]")
    for i in range(boxes_shape[0]):
        display_str_list = ()
        if display_str_list_list:
            display_str_list = display_str_list_list[i]
        draw_bounding_box_on_image(
            image,
            boxes[i, 0],
            boxes[i, 1],
            boxes[i, 2],
            boxes[i, 3],
            color,
            thickness,
            display_str_list,
        )


def _visualize_boxes(image, boxes, classes, scores, category_index, **kwargs):
    return visualize_boxes_and_labels_on_image_array(
        image, boxes, classes, scores, category_index=category_index, **kwargs
    )


def _visualize_boxes_and_masks(
    image, boxes, classes, scores, masks, category_index, **kwargs
):
    return visualize_boxes_and_labels_on_image_array(
        image,
        boxes,
        classes,
        scores,
        category_index=category_index,
        instance_masks=masks,
        **kwargs
    )


def _visualize_boxes_and_keypoints(
    image, boxes, classes, scores, keypoints, category_index, **kwargs
):
    return visualize_boxes_and_labels_on_image_array(
        image,
        boxes,
        classes,
        scores,
        category_index=category_index,
        keypoints=keypoints,
        **kwargs
    )


def _visualize_boxes_and_masks_and_keypoints(
    image, boxes, classes, scores, masks, keypoints, category_index, **kwargs
):
    return visualize_boxes_and_labels_on_image_array(
        image,
        boxes,
        classes,
        scores,
        category_index=category_index,
        instance_masks=masks,
        keypoints=keypoints,
        **kwargs
    )


def draw_keypoints_on_image_array(
    image, keypoints, color="red", radius=2, use_normalized_coordinates=True
):
    """Draws keypoints on an image (numpy array).

    Args:
      image: a numpy array with shape [height, width, 3].
      keypoints: a numpy array with shape [num_keypoints, 2].
      color: color to draw the keypoints with. Default is red.
      radius: keypoint radius. Default value is 2.
      use_normalized_coordinates: if True (default), treat keypoint values as
        relative to the image.  Otherwise treat them as absolute.
    """
    image_pil = Image.fromarray(np.uint8(image)).convert("RGB")
    draw_keypoints_on_image(
        image_pil, keypoints, color, radius, use_normalized_coordinates
    )
    np.copyto(image, np.array(image_pil))


def draw_keypoints_on_image(
    image, keypoints, color="red", radius=2, use_normalized_coordinates=True
):
    """Draws keypoints on an image.

    Args:
      image: a PIL.Image object.
      keypoints: a numpy array with shape [num_keypoints, 2].
      color: color to draw the keypoints with. Default is red.
      radius: keypoint radius. Default value is 2.
      use_normalized_coordinates: if True (default), treat keypoint values as
        relative to the image.  Otherwise treat them as absolute.
    """
    draw = ImageDraw.Draw(image)
    im_width, im_height = image.size
    keypoints_x = [k[1] for k in keypoints]
    keypoints_y = [k[0] for k in keypoints]
    if use_normalized_coordinates:
        keypoints_x = tuple([im_width * x for x in keypoints_x])
        keypoints_y = tuple([im_height * y for y in keypoints_y])
    for keypoint_x, keypoint_y in zip(keypoints_x, keypoints_y):
        draw.ellipse(
            [
                (keypoint_x - radius, keypoint_y - radius),
                (keypoint_x + radius, keypoint_y + radius),
            ],
            outline=color,
            fill=color,
        )


def draw_mask_on_image_array(image, mask, color="red", alpha=0.4):
    """Draws mask on an image.

    Args:
      image: uint8 numpy array with shape (img_height, img_height, 3)
      mask: a uint8 numpy array of shape (img_height, img_height) with
        values between either 0 or 1.
      color: color to draw the keypoints with. Default is red.
      alpha: transparency value between 0 and 1. (default: 0.4)

    Raises:
      ValueError: On incorrect data type for image or masks.
    """
    if image.dtype != np.uint8:
        raise ValueError("`image` not of type np.uint8")
    if mask.dtype != np.uint8:
        raise ValueError("`mask` not of type np.uint8")
    if np.any(np.logical_and(mask != 1, mask != 0)):
        raise ValueError("`mask` elements should be in [0, 1]")
    if image.shape[:2] != mask.shape:
        raise ValueError(
            "The image has spatial dimensions %s but the mask has "
            "dimensions %s" % (image.shape[:2], mask.shape)
        )
    rgb = ImageColor.getrgb(color)
    pil_image = Image.fromarray(image)

    solid_color = np.expand_dims(np.ones_like(mask), axis=2) * np.reshape(
        list(rgb), [1, 1, 3]
    )
    pil_solid_color = Image.fromarray(np.uint8(solid_color)).convert("RGBA")
    pil_mask = Image.fromarray(np.uint8(255.0 * alpha * mask)).convert("L")
    pil_image = Image.composite(pil_solid_color, pil_image, pil_mask)
    np.copyto(image, np.array(pil_image.convert("RGB")))


def visualize_boxes_and_labels_on_image_array(
    image,
    boxes,
    classes,
    scores,
    label_names,
    instance_masks=None,
    instance_boundaries=None,
    keypoints=None,
    use_normalized_coordinates=False,
    max_boxes_to_draw=20,
    min_score_thresh=0.5,
    agnostic_mode=False,
    line_thickness=4,
    groundtruth_box_visualization_color="black",
    skip_scores=False,
    skip_labels=False,
):
    """Overlay labeled boxes on an image with formatted scores and label names.

    This function groups boxes that correspond to the same location
    and creates a display string for each detection and overlays these
    on the image. Note that this function modifies the image in place, and returns
    that same image.

    Args:
      image: uint8 numpy array with shape (img_height, img_width, 3)
      boxes: a numpy array of shape [N, 4]
      classes: a numpy array of shape [N]. Note that class indices are 1-based,
        and match the keys in the label map.
      scores: a numpy array of shape [N] or None.  If scores=None, then
        this function assumes that the boxes to be plotted are groundtruth
        boxes and plot all boxes as black with no classes or scores.
      label_names: a list containing label names.
      instance_masks: a numpy array of shape [N, image_height, image_width] with
        values ranging between 0 and 1, can be None.
      instance_boundaries: a numpy array of shape [N, image_height, image_width]
        with values ranging between 0 and 1, can be None.
      keypoints: a numpy array of shape [N, num_keypoints, 2], can
        be None
      use_normalized_coordinates: whether boxes is to be interpreted as
        normalized coordinates or not.
      max_boxes_to_draw: maximum number of boxes to visualize.  If None, draw
        all boxes.
      min_score_thresh: minimum score threshold for a box to be visualized
      agnostic_mode: boolean (default: False) controlling whether to evaluate in
        class-agnostic mode or not.  This mode will display scores but ignore
        classes.
      line_thickness: integer (default: 4) controlling line width of the boxes.
      groundtruth_box_visualization_color: box color for visualizing groundtruth
        boxes
      skip_scores: whether to skip score when drawing a single detection
      skip_labels: whether to skip label when drawing a single detection

    Returns:
      uint8 numpy array with shape (img_height, img_width, 3) with overlaid boxes.
    """
    # Create a display string (and color) for every box location, group any boxes
    # that correspond to the same location.
    box_to_display_str_map = collections.defaultdict(list)
    box_to_color_map = collections.defaultdict(str)
    box_to_instance_masks_map = {}
    box_to_instance_boundaries_map = {}
    box_to_keypoints_map = collections.defaultdict(list)
    if not max_boxes_to_draw:
        max_boxes_to_draw = boxes.shape[0]
    for i in range(min(max_boxes_to_draw, boxes.shape[0])):
        if scores is None or scores[i] > min_score_thresh:
            box = tuple(boxes[i].tolist())
            if instance_masks is not None:
                box_to_instance_masks_map[box] = instance_masks[i]
            if instance_boundaries is not None:
                box_to_instance_boundaries_map[box] = instance_boundaries[i]
            if keypoints is not None:
                box_to_keypoints_map[box].extend(keypoints[i])
            if scores is None:
                box_to_color_map[box] = groundtruth_box_visualization_color
            else:
                display_str = ""
                if not skip_labels:
                    if not agnostic_mode:
                        if len(label_names) > classes[i]:
                            class_name = label_names[classes[i]]
                        else:
                            class_name = "N/A"
                        display_str = str(class_name)
                if not skip_scores:
                    if not display_str:
                        display_str = "{}%".format(int(100 * scores[i]))
                    else:
                        display_str = "{}: {}%".format(
                            display_str, int(100 * scores[i])
                        )
                box_to_display_str_map[box].append(display_str)
                if agnostic_mode:
                    box_to_color_map[box] = "DarkOrange"
                else:
                    box_to_color_map[box] = STANDARD_COLORS[
                        classes[i] % len(STANDARD_COLORS)
                    ]

    # Draw all boxes onto image.
    for box, color in box_to_color_map.items():
        ymin, xmin, ymax, xmax = box
        if instance_masks is not None:
            draw_mask_on_image_array(image, box_to_instance_masks_map[box], color=color)
        if instance_boundaries is not None:
            draw_mask_on_image_array(
                image, box_to_instance_boundaries_map[box], color="red", alpha=1.0
            )
        draw_bounding_box_on_image_array(
            image,
            ymin,
            xmin,
            ymax,
            xmax,
            color=color,
            thickness=line_thickness,
            display_str_list=box_to_display_str_map[box],
            use_normalized_coordinates=use_normalized_coordinates,
        )
        if keypoints is not None:
            draw_keypoints_on_image_array(
                image,
                box_to_keypoints_map[box],
                color=color,
                radius=line_thickness / 2,
                use_normalized_coordinates=use_normalized_coordinates,
            )

    return image



import operator
from typing import Optional

import cachetools

from robotoff import settings
from robotoff.products import ProductDataset
from robotoff.taxonomy import TaxonomyType, get_taxonomy
from robotoff.utils import dump_json, dump_text, http_session, load_json, text_file_iter
from robotoff.utils.cache import CachedStore


@cachetools.cached(cachetools.LRUCache(maxsize=1))
def get_brand_prefix() -> set[tuple[str, str]]:
    """Get a set of brand prefix tuples found in Open Food Facts databases.

    Each tuple has the format (brand_tag, prefix) where prefix is a digit with
    13 elements (EAN-13).
    """
    return set(tuple(x) for x in load_json(settings.BRAND_PREFIX_PATH, compressed=True))  # type: ignore


def get_brand_blacklist() -> set[str]:
    return set(text_file_iter(settings.OCR_TAXONOMY_BRANDS_BLACKLIST_PATH))


def generate_barcode_prefix(barcode: str) -> str:
    if len(barcode) == 13:
        prefix = 7
        return barcode[:prefix] + "x" * (len(barcode) - prefix)

    raise ValueError(f"barcode prefix only works on EAN-13 barcode (here: {barcode})")


def compute_brand_prefix(
    product_dataset: ProductDataset, threshold: Optional[int] = None
) -> dict[tuple[str, str], int]:
    count: dict[tuple[str, str], int] = {}

    for product in (
        product_dataset.stream()
        .filter_nonempty_tag_field("brands_tags")
        .filter_nonempty_text_field("code")
    ):
        brand_tags = set(x for x in product["brands_tags"] if x)
        barcode = product["code"]

        if len(barcode) == 13:
            barcode_prefix = generate_barcode_prefix(barcode)

            for brand_tag in brand_tags:
                key = (brand_tag, barcode_prefix)
                count.setdefault(key, 0)
                count[key] += 1

    if threshold:
        for key in list(count.keys()):
            if count[key] < threshold:
                count.pop(key)

    return count


def save_brand_prefix(count_threshold: int = 5):
    product_dataset = ProductDataset(settings.JSONL_DATASET_PATH)
    counts = compute_brand_prefix(product_dataset, threshold=count_threshold)
    brand_prefixes = list(counts.keys())
    dump_json(settings.BRAND_PREFIX_PATH, brand_prefixes, compressed=True)


def keep_brand_from_taxonomy(
    brand_tag: str,
    brand: str,
    min_length: Optional[int] = None,
    blacklisted_brands: Optional[set[str]] = None,
) -> bool:
    if brand.isdigit():
        return False

    if min_length and len(brand) < min_length:
        return False

    if blacklisted_brands is not None and brand_tag in blacklisted_brands:
        return False

    return True


def generate_brand_list(
    threshold: int,
    min_length: Optional[int] = None,
    blacklisted_brands: Optional[set[str]] = None,
) -> list[tuple[str, str]]:
    min_length = min_length or 0
    brand_taxonomy = get_taxonomy(TaxonomyType.brand.name)
    url = settings.BaseURLProvider().get() + "/brands.json"
    brand_count_list = http_session.get(url).json()["tags"]

    brand_count = {tag["id"]: tag for tag in brand_count_list}
    brand_list = []

    for node in brand_taxonomy.iter_nodes():
        key = node.id
        name = node.names["en"]

        if key.startswith("en:"):
            key = key[3:]

        if (
            keep_brand_from_taxonomy(key, name, min_length, blacklisted_brands)
            and brand_count.get(key, {}).get("products", 0) >= threshold
        ):
            brand_list.append((key, name))

    return sorted(brand_list, key=operator.itemgetter(0))


def dump_taxonomy_brands(
    threshold: int,
    min_length: Optional[int] = None,
    blacklisted_brands: Optional[set[str]] = None,
):
    filtered_brands = generate_brand_list(threshold, min_length, blacklisted_brands)
    line_iter = ("{}||{}".format(key, name) for key, name in filtered_brands)
    dump_text(settings.OCR_TAXONOMY_BRANDS_PATH, line_iter)


def in_barcode_range(
    brand_prefix: set[tuple[str, str]], brand_tag: str, barcode: str
) -> bool:
    """Check that the insight barcode is in the range of the detected
    brand barcode range.
    Return True if the check passes, False otherwise
    """
    if len(barcode) == 13:
        barcode_prefix = generate_barcode_prefix(barcode)
        key = (brand_tag, barcode_prefix)

        if key not in brand_prefix:
            return False

    return True


BRAND_BLACKLIST_STORE = CachedStore(
    fetch_func=get_brand_blacklist, expiration_interval=None
)

if __name__ == "__main__":
    blacklisted_brands = get_brand_blacklist()
    dump_taxonomy_brands(
        threshold=settings.BRAND_MATCHING_MIN_COUNT,
        min_length=settings.BRAND_MATCHING_MIN_LENGTH,
        blacklisted_brands=blacklisted_brands,
    )



import datetime
import uuid
from typing import Optional

from robotoff import settings
from robotoff.insights.dataclass import InsightType
from robotoff.models import Prediction, ProductInsight, with_db
from robotoff.off import get_server_type
from robotoff.products import (
    DBProductStore,
    get_image_id,
    get_product_store,
    has_nutrition_image,
    is_nutrition_image,
    is_valid_image,
)
from robotoff.types import PredictionType
from robotoff.utils import get_logger
from robotoff.utils.types import JSONType

logger = get_logger(__name__)

FIBER_QUALITY_FACET_NAME = "en:missing-nutrition-facts-fibers-present-on-photos"
FIBER_NUTRITION_QUALITY_FACET_NAME = (
    "en:missing-nutrition-facts-fibers-present-on-nutrition-photos"
)


def generate_quality_facets():
    generate_fiber_quality_facet()


@with_db
def generate_fiber_quality_facet():
    product_store: DBProductStore = get_product_store()
    collection = product_store.collection
    added = 0
    seen_set: set[str] = set()

    for prediction in (
        Prediction.select(Prediction.barcode, Prediction.source_image)
        .where(
            Prediction.type == PredictionType.nutrient_mention.name,
            Prediction.data["mentions"].contains("fiber"),
            Prediction.source_image.is_null(False),
        )
        .iterator()
    ):
        barcode = prediction.barcode

        if barcode in seen_set:
            continue

        product = product_store.get_product(
            barcode, ["nutriments", "data_quality_tags", "images"]
        )

        if product is None:
            continue

        nutriments = product.get("nutriments", {})
        data_quality_tags = product.get("data_quality_tags", {})
        images = product.get("images", {})

        if (
            not is_valid_image(images, prediction.source_image)
            or "fiber" in nutriments
            or "fiber_prepared" in nutriments
        ):
            continue

        facets = []

        if FIBER_QUALITY_FACET_NAME not in data_quality_tags:
            facets.append(FIBER_QUALITY_FACET_NAME)

        if (
            FIBER_NUTRITION_QUALITY_FACET_NAME not in data_quality_tags
            and is_nutrition_image(images, prediction.source_image)
        ):
            facets.append(FIBER_NUTRITION_QUALITY_FACET_NAME)

        if not facets:
            continue

        logger.info("Adding facets to {}: {}".format(barcode, facets))
        seen_set.add(barcode)
        added += 1
        collection.update_one(
            {"code": barcode},
            {
                "$push": {
                    "data_quality_tags": {"$each": facets},
                    "data_quality_warnings_tags": {"$each": facets},
                }
            },
        )
    logger.info("Fiber quality facets added on {} products".format(added))


def get_image_orientation(barcode: str, image_id: str) -> Optional[int]:
    for prediction in (
        Prediction.select(Prediction.data, Prediction.source_image)
        .where(
            Prediction.barcode == barcode,
            Prediction.type == PredictionType.image_orientation.name,
            Prediction.server_domain == settings.OFF_SERVER_DOMAIN,
            Prediction.source_image.is_null(False),
        )
        .iterator()
    ):
        prediction_image_id = get_image_id(prediction.source_image)  # type: ignore

        if image_id is not None and prediction_image_id == image_id:
            return prediction.data.get("rotation")

    return None


def generate_nutrition_image_insights():
    logger.info("Starting nutrition image insight generation")
    logger.info("Deleting previous nutrition image insights...")
    deleted = (
        ProductInsight.delete()
        .where(
            ProductInsight.annotation.is_null(),
            ProductInsight.type == InsightType.nutrition_image.name,
            ProductInsight.server_domain == settings.OFF_SERVER_DOMAIN,
        )
        .execute()
    )
    logger.info("{} insights deleted".format(deleted))
    product_store: DBProductStore = get_product_store()
    added = 0
    seen_set: set[str] = set()

    prediction: Prediction
    for prediction in (
        Prediction.select()
        .where(Prediction.type == PredictionType.nutrient_mention.name)
        .order_by(Prediction.source_image.desc())
        .iterator()
    ):
        barcode = prediction.barcode

        if barcode in seen_set:
            continue

        mentions = prediction.data["mentions"]
        nutrition_image_langs = find_nutrition_image_lang(mentions)

        if not nutrition_image_langs:
            continue

        image_id = get_image_id(prediction.source_image)
        rotation = get_image_orientation(barcode, image_id)

        if rotation is None:
            continue

        product = product_store.get_product(barcode, ["images"])

        if product is None:
            continue

        images = product.get("images", {})

        if not has_nutrition_image(images):
            for lang in nutrition_image_langs:
                if not (
                    Prediction.select()
                    .where(
                        Prediction.type == PredictionType.nutrition_image.name,
                        Prediction.barcode == barcode,
                        Prediction.value_tag == lang,
                        Prediction.server_domain == settings.OFF_SERVER_DOMAIN,
                    )
                    .count()
                ):
                    ProductInsight.create(
                        id=str(uuid.uuid4()),
                        barcode=prediction.barcode,
                        type=InsightType.nutrition_image.name,
                        value_tag=lang,
                        timestamp=datetime.datetime.utcnow(),
                        source_image=prediction.source_image,
                        server_domain=prediction.server_domain,
                        server_type=get_server_type(prediction.server_domain).name,
                        automatic_processing=False,
                        data={
                            "from_prediction": str(prediction.id),
                            "languages": nutrition_image_langs,
                            "rotation": rotation or None,
                        },
                    )
                    added += 1

    logger.info("Added: {}".format(added))


def find_nutrition_image_lang(mentions: JSONType, min_count: int = 4) -> list[str]:
    nutrient_languages = find_nutrition_image_nutrient_languages(mentions)

    lang_count: dict[str, int] = {}
    for _, langs in nutrient_languages.items():
        for lang, count in langs.items():
            lang_count.setdefault(lang, 0)
            lang_count[lang] += count

    return [lang for lang, count in lang_count.items() if count >= min_count]


def find_nutrition_image_nutrient_languages(
    mentions: JSONType,
) -> dict[str, dict[str, int]]:
    languages: dict[str, dict[str, int]] = {}
    for nutrient, matches in mentions.items():
        seen_lang: set[str] = set()

        for match in matches:
            for lang in match.get("languages", []):
                if lang not in seen_lang:
                    languages.setdefault(nutrient, {})
                    nutrient_languages = languages[nutrient]
                    nutrient_languages.setdefault(lang, 0)
                    nutrient_languages[lang] += 1
                    seen_lang.add(lang)

    return languages



import abc
import datetime
import itertools
import operator
import uuid
from pathlib import Path
from typing import Any, Iterable, Iterator, Optional, Type

from playhouse.shortcuts import model_to_dict

from robotoff import settings
from robotoff.brands import get_brand_prefix, in_barcode_range
from robotoff.insights.dataclass import InsightType
from robotoff.insights.normalize import normalize_emb_code
from robotoff.models import Prediction as PredictionModel
from robotoff.models import ProductInsight, batch_insert
from robotoff.off import get_server_type
from robotoff.prediction.types import Prediction
from robotoff.products import (
    DBProductStore,
    Product,
    get_image_id,
    get_product_store,
    is_valid_image,
)
from robotoff.redis import Lock, LockedResourceException
from robotoff.taxonomy import (
    Taxonomy,
    TaxonomyType,
    get_taxonomy,
    match_taxonomized_value,
)
from robotoff.types import PredictionType
from robotoff.utils import get_logger, text_file_iter
from robotoff.utils.cache import CachedStore

logger = get_logger(__name__)


def load_authorized_labels() -> set[str]:
    return set(text_file_iter(settings.OCR_LABEL_WHITELIST_DATA_PATH))


AUTHORIZED_LABELS_STORE = CachedStore(load_authorized_labels, expiration_interval=None)


def is_selected_image(images: dict[str, Any], image_id: str) -> bool:
    """Return True if the image referenced by `image_id` is selected as a
    front, ingredients, nutrition or packaging image in any language.

    :param images: The image dict as stored in MongoDB.
    :param image_id: The image ID to compare, must be a digit.
    """
    for key_prefix in ("front", "ingredients", "nutrition", "packaging"):
        for key, image in images.items():
            if key.startswith(key_prefix) and image["imgid"] == image_id:
                return True

    return False


def is_recent_image(
    images: dict[str, Any], image_id: str, max_timedelta: datetime.timedelta
) -> bool:
    """Return True if the image referenced by `image_id` is less than
    `max_timedelta` older than the most recent image, return False otherwise.

    Only "raw" images (images identified with digits) are compared to the
    provided `image_id`.

    :param images: The image dict as stored in MongoDB.
    :param image_id: The image ID to compare, must be a digit.
    :param max_timedelta: The maximum interval between the upload datetime of
    the most recent image and the provided image.
    """
    image_datetime = datetime.datetime.utcfromtimestamp(
        int(images[image_id]["uploaded_t"])
    )
    remaining_datetimes = []
    for key, image_meta in images.items():
        if key.isdigit() and key != image_id:
            remaining_datetimes.append(
                datetime.datetime.utcfromtimestamp(int(image_meta["uploaded_t"]))
            )

    for upload_datetime in remaining_datetimes:
        if upload_datetime - image_datetime > max_timedelta:
            logger.debug("More recent image: %s > %s", upload_datetime, image_datetime)
            return False

    return True


def is_valid_insight_image(images: dict[str, Any], source_image: Optional[str]):
    """Return True if the source image is valid for insight generation,
    i.e. the image ID is a digit and is referenced in `images`.

    If `source_image` is None, we always consider the insight as valid.

    :param images: The image dict as stored in MongoDB.
    :param source_image: The insight source image, should be the path of the
    image path or None.
    """
    if source_image is None:
        return True

    image_id = Path(source_image).stem
    return image_id.isdigit() and image_id in images


def is_trustworthy_insight_image(
    images: dict[str, Any],
    source_image: Optional[str],
    max_timedelta: datetime.timedelta = settings.IMAGE_MAX_TIMEDELTA,
):
    """Return True if the source image is trustworthy for insight generation,
      - the image ID is a digit and is referenced in `images`
      - the image is either selected or recent enough

    If `source_image` is None, we always consider the insight as trustworthy.
    Insights considered as trustworthy can have automatic_processing = True.

    :param images: The image dict as stored in MongoDB.
    :param source_image: The insight source image, should be the path of the
    image path or None.
    :param max_timedelta: Maximum timedelta between most recent image and
    source image, default settings.IMAGE_MAX_TIMEDELTA.
    """
    if source_image is None:
        return True

    image_id = Path(source_image).stem

    if not image_id.isdigit() or image_id not in images:
        return False

    return is_selected_image(images, image_id) or is_recent_image(
        images, image_id, max_timedelta
    )


def get_existing_insight(
    insight_type: InsightType, barcode: str, server_domain: str
) -> list[ProductInsight]:
    """Get all insights for specific product and `insight_type`."""
    return list(
        ProductInsight.select().where(
            ProductInsight.type == insight_type.name,
            ProductInsight.barcode == barcode,
            ProductInsight.server_domain == server_domain,
        )
    )


def is_reserved_barcode(barcode: str) -> bool:
    if barcode.startswith("0"):
        barcode = barcode[1:]

    return barcode.startswith("2")


def sort_predictions(predictions: Iterable[Prediction]) -> list[Prediction]:
    """Sort predictions by priority, using as keys:
    - priority, specified by data["priority"], prediction with lowest priority
    values (high priority) come first
    - source image upload datetime (most recent first): images IDs are
    auto-incremented integers, so the most recent images have the highest IDs.
    Images with `source_image = None` have a lower priority that images with a
    source image.
    - predictor, predictions with predictor value have higher priority

    :param predictions: The predictions to sort
    :return: Sorted predictions
    """
    return sorted(
        predictions,
        key=lambda prediction: (
            prediction.data.get("priority", 1),
            -int(get_image_id(prediction.source_image) or 0)
            if prediction.source_image
            else 0,
            # hack to set a higher priority to prediction with a predictor value
            prediction.predictor or "z",
        ),
    )


def sort_candidates(candidates: Iterable[ProductInsight]) -> list[ProductInsight]:
    """Sort candidates by priority, using as keys:

    - priority, specified by data["priority"], candidate with lowest priority
      values (high priority) come first
    - source image upload datetime (most recent first): images IDs are
      auto-incremented integers, so the most recent images have the highest IDs.
      Images with `source_image = None` have a lower priority that images with a
      source image.
    - automatic processing status: candidates that are automatically
      processable have higher priority

    This function should be used to make sure most important candidates are
    looked into first in `get_insight_update`. Note that the sorting keys are
    a superset of those used in `sort_predictions`.

    :param candidates: The candidates to sort
    :return: Sorted candidates
    """
    return sorted(
        candidates,
        key=lambda candidate: (
            candidate.data.get("priority", 1),
            -int(get_image_id(candidate.source_image) or 0)
            if candidate.source_image
            else 0,
            # automatically processable insights come first
            -int(candidate.automatic_processing),
            # hack to set a higher priority to prediction with a predictor value
            candidate.predictor or "z",
        ),
    )


def select_deepest_taxonomized_candidates(
    candidates: list[Prediction], taxonomy: Taxonomy
):
    """Filter predictions to only keep the deepest items in the taxonomy.

    For instance, for a list of category predictions, the prediction with
    `value_tag` 'en:meat' will be removed if a prediction with `value_tag`
    'en:pork' is in the `candidates` list.

    :param candidates: The list of candidates to filter
    :param taxonomy: The taxonomy to use
    """
    value_tags = set()

    for candidate in candidates:
        if candidate.value_tag is None:
            logger.warning("Unexpected None `value_tag` (candidate: %s)", candidate)
        else:
            value_tags.add(candidate.value_tag)

    nodes = [taxonomy[node] for node in value_tags if node in taxonomy]
    selected_node_ids = set(node.id for node in taxonomy.find_deepest_nodes(nodes))
    return [
        candidate
        for candidate in candidates
        if candidate.value_tag in selected_node_ids
    ]


class InsightImporter(metaclass=abc.ABCMeta):
    """Abstract class for all insight importers."""

    @staticmethod
    @abc.abstractmethod
    def get_type() -> InsightType:
        """Return the type of generated insights.

        This method must be implemented in subclasses."""
        pass

    @staticmethod
    @abc.abstractmethod
    def get_required_prediction_types() -> set[PredictionType]:
        """Return the prediction types that are necessary to generate the
        insight type.

        For most insight types a set of a single element will be returned, but
        more complex insight types require several prediction types.
        This method must be implemented in subclasses.
        """
        pass

    @classmethod
    def import_insights(
        cls,
        barcode: str,
        predictions: list[Prediction],
        server_domain: str,
        product_store: DBProductStore,
    ) -> int:
        """Import insights, this is the main method.

        :return: the number of insights that were imported.
        """
        required_prediction_types = cls.get_required_prediction_types()
        for prediction in predictions:
            if prediction.type not in required_prediction_types:
                raise ValueError(f"unexpected prediction type: '{prediction.type}'")

        if (
            len(
                prediction_barcodes := set(
                    prediction.barcode for prediction in predictions
                )
            )
            > 1
        ):
            raise ValueError(
                f"predictions for more than 1 product were provided: {prediction_barcodes}"
            )

        inserts = 0
        to_create, to_update, to_delete = cls.generate_insights(
            barcode, predictions, server_domain, product_store
        )
        if to_delete:
            to_delete_ids = [insight.id for insight in to_delete]
            logger.info("Deleting %s insights", len(to_delete_ids))
            ProductInsight.delete().where(
                ProductInsight.id.in_(to_delete_ids)
            ).execute()
        if to_create:
            inserts += batch_insert(
                ProductInsight,
                (model_to_dict(insight) for insight in to_create),
                50,
            )

        for insight in to_update:
            insight.save()

        return inserts

    @classmethod
    def generate_insights(
        cls,
        barcode: str,
        predictions: list[Prediction],
        server_domain: str,
        product_store: DBProductStore,
    ) -> tuple[list[ProductInsight], list[ProductInsight], list[ProductInsight]]:
        """Given a list of predictions, yield tuples of ProductInsight to
        create, update and delete.

        It calls the `generate_candidates` method, specific to each insight type
        (and implemented in sub-classes).
        """
        timestamp = datetime.datetime.utcnow()
        server_type = get_server_type(server_domain).name

        product = product_store[barcode]
        references = get_existing_insight(cls.get_type(), barcode, server_domain)

        if product is None:
            logger.info(
                f"Product {barcode} not found in DB, deleting existing insights"
            )
            return [], [], references

        predictions = sort_predictions(predictions)
        candidates = [
            candidate
            for candidate in cls.generate_candidates(product, predictions)
            if is_valid_insight_image(product.images, candidate.source_image)
        ]
        for candidate in candidates:
            if candidate.automatic_processing is None:
                logger.warning(
                    "Insight with automatic_processing=None: %s", candidate.__data__
                )
                candidate.automatic_processing = False

            if not is_trustworthy_insight_image(product.images, candidate.source_image):
                # Don't process automatically if the insight image is not
                # trustworthy (too old and not selected)
                candidate.automatic_processing = False
            if candidate.data.get("is_annotation"):
                username = candidate.data.get("username")
                if username:
                    # logo annotation by a user
                    candidate.username = username
                # Note: we could add vote annotation for anonymous user,
                # but it should be done outside this loop. It's not yet implemented

        to_create, to_update, to_delete = cls.get_insight_update(candidates, references)

        for insight in to_create:
            cls.add_fields(insight, product, timestamp, server_domain, server_type)

        for insight, reference_insight in to_update:
            # Keep `reference_insight` in DB (as the value/value_tag/source_image is the same),
            # but update information from `insight`.
            # This way, we don't unnecessarily insert/delete rows in ProductInsight table
            # and we keep associated votes
            cls.update_fields(insight, reference_insight, product, timestamp)

        return (
            to_create,
            [reference_insight for (_, reference_insight) in to_update],
            to_delete,
        )

    @classmethod
    @abc.abstractmethod
    def generate_candidates(
        cls,
        product: Product,
        predictions: list[Prediction],
    ) -> Iterator[ProductInsight]:
        """From a list of Predictions associated with a product, yield
        candidate ProductInsights for import.

        The types of all Predictions must be a subset of the required types
        available by calling `InsightImporter.get_required_prediction_types`.
        This method must be implemented in subclasses.

        :param product: The Product associated with the Predictions
        :param predictions: The list of predictions for the product of type
        :yield: candidate ProductInsight
        """
        pass

    @classmethod
    def get_insight_update(
        cls, candidates: list[ProductInsight], reference_insights: list[ProductInsight]
    ) -> tuple[
        list[ProductInsight],
        list[tuple[ProductInsight, ProductInsight]],
        list[ProductInsight],
    ]:
        """Return a tuple containing:
        - a list of ProductInsight to create
        - a list of ProductInsight to update, as (insight, reference_insight)
          tuples, where `insight` is the candidate and `reference_insight` is
          the insight already in DB
        - a list of ProductInsight to delete

        :param candidates: candidate predictions
        :param reference_insights: existing insights of this type and product
        """
        to_create_or_update: list[tuple[ProductInsight, Optional[ProductInsight]]] = []
        # Keep already annotated insights in DB
        to_keep_ids = set(
            reference.id
            for reference in reference_insights
            # Don't delete already annotated insights
            if reference.annotation is not None
            # Don't overwrite an insight that is going to be applied
            # automatically soon
            or reference.automatic_processing is True
        )
        for candidate in sort_candidates(candidates):
            # if match is True, candidate conflicts with existing insight,
            # keeping existing insight and discarding candidate
            match = any(
                cls.is_conflicting_insight(candidate, reference_insight)
                for reference_insight in reference_insights
                if reference_insight.annotation is not None
                # Don't overwrite an insight that is going to be applied
                # automatically soon
                or reference_insight.automatic_processing is True
            )

            if not match:
                for selected, _ in to_create_or_update:
                    if cls.is_conflicting_insight(candidate, selected):
                        # Don't import candidate if it conflicts with an
                        # already selected candidate
                        break
                else:
                    mapping_ref_insight = None
                    # In order for the voting system to work, we map insights to create
                    # to existing insights with the same value/value_tag/source_image.
                    # This way, we don't loose vote information.
                    for reference_insight in reference_insights:
                        if (
                            reference_insight.annotation is None
                            and cls.is_conflicting_insight(candidate, reference_insight)
                            # only map to existing insight if the source image is the same,
                            # otherwise create a new insight
                            and candidate.source_image == reference_insight.source_image
                        ):
                            mapping_ref_insight = reference_insight
                            to_keep_ids.add(reference_insight.id)
                            break

                    # If mapping_ref_insight is None, a new insight is created in DB,
                    # Otherwise the reference insight is updated with candidate
                    # information
                    to_create_or_update.append((candidate, mapping_ref_insight))

        to_delete = [
            insight for insight in reference_insights if insight.id not in to_keep_ids
        ]
        to_create = [
            insight
            for insight, ref_insight in to_create_or_update
            if ref_insight is None
        ]
        to_update = [
            (insight, ref_insight)
            for insight, ref_insight in to_create_or_update
            if ref_insight is not None
        ]
        return to_create, to_update, to_delete

    @classmethod
    @abc.abstractmethod
    def is_conflicting_insight(
        cls, candidate: ProductInsight, reference: ProductInsight
    ) -> bool:
        """Return True if a candidate ProductInsight conflicts with an
        existing or another candidate insight, in which case the candidate
        insight won't be imported.

        :param candidate: The candidate ProductInsight to import
        :param reference: A ProductInsight, either another candidate or an
        insight that exists in DB
        """
        pass

    @classmethod
    def add_fields(
        cls,
        insight: ProductInsight,
        product: Product,
        timestamp: datetime.datetime,
        server_domain: str,
        server_type: str,
    ):
        """Add mandatory insight fields."""
        barcode = insight.barcode
        insight.reserved_barcode = is_reserved_barcode(barcode)
        insight.server_domain = server_domain
        insight.server_type = server_type
        insight.id = str(uuid.uuid4())
        insight.timestamp = timestamp
        insight.countries = product.countries_tags
        insight.brands = product.brands_tags
        insight.n_votes = 0
        insight.unique_scans_n = product.unique_scans_n

        if insight.automatic_processing:
            insight.process_after = timestamp + datetime.timedelta(
                minutes=settings.INSIGHT_AUTOMATIC_PROCESSING_WAIT
            )

        cls.add_optional_fields(insight, product)

    @classmethod
    def update_fields(
        cls,
        insight: ProductInsight,
        reference_insight: ProductInsight,
        product: Product,
        timestamp: datetime.datetime,
    ):
        """Update `reference_insight` fields with data from `insight`.

        This is used to refresh `reference_insight` with potentially fresher
        information, while avoiding row deletion/insertion each time
        `import_insights` is called.
        """
        cls.add_fields(
            insight,
            product,
            timestamp,
            reference_insight.server_domain,
            reference_insight.server_type,
        )

        for field_name in (
            key
            for key in insight.__data__.keys()
            if key not in ("id", "barcode", "type", "server_domain", "server_type")
        ):
            if getattr(insight, field_name) != getattr(reference_insight, field_name):
                setattr(reference_insight, field_name, getattr(insight, field_name))

    @classmethod
    def add_optional_fields(  # noqa: B027
        cls, insight: ProductInsight, product: Product
    ):
        """Overwrite this method in children classes to add optional fields.

        The `campaign` field should be populated here.

        :param insight: the ProductInsight
        :param product: the associated Product
        """
        pass


class PackagerCodeInsightImporter(InsightImporter):
    @staticmethod
    def get_type() -> InsightType:
        return InsightType.packager_code

    @staticmethod
    def get_required_prediction_types() -> set[PredictionType]:
        return {PredictionType.packager_code}

    @classmethod
    def is_conflicting_insight(
        cls, candidate: ProductInsight, reference: ProductInsight
    ) -> bool:
        return candidate.value == reference.value

    @staticmethod
    def is_prediction_valid(
        product: Product,
        emb_code: str,
    ) -> bool:
        existing_codes = [normalize_emb_code(c) for c in product.emb_codes_tags]
        normalized_code = normalize_emb_code(emb_code)

        return normalized_code not in existing_codes

    @classmethod
    def generate_candidates(
        cls,
        product: Product,
        predictions: list[Prediction],
    ) -> Iterator[ProductInsight]:
        yield from (
            ProductInsight(**prediction.to_dict())
            for prediction in predictions
            if cls.is_prediction_valid(product, prediction.value)  # type: ignore
        )


class LabelInsightImporter(InsightImporter):
    @staticmethod
    def get_type() -> InsightType:
        return InsightType.label

    @staticmethod
    def get_required_prediction_types() -> set[PredictionType]:
        return {PredictionType.label}

    @classmethod
    def is_conflicting_insight(
        cls, candidate: ProductInsight, reference: ProductInsight
    ) -> bool:
        return candidate.value_tag == reference.value_tag or cls.is_parent_label(
            candidate.value_tag, {reference.value_tag}  # type: ignore
        )

    @staticmethod
    def is_prediction_valid(product: Product, tag: str) -> bool:
        return not (
            tag in product.labels_tags
            or LabelInsightImporter.is_parent_label(tag, set(product.labels_tags))
        )

    @classmethod
    def is_parent_label(cls, tag: str, to_check_labels: set[str]) -> bool:
        # Check that the predicted label is not a parent of a
        # current/already predicted label
        return get_taxonomy(InsightType.label.name).is_parent_of_any(
            tag, to_check_labels, raises=False
        )

    @classmethod
    def generate_candidates(
        cls,
        product: Product,
        predictions: list[Prediction],
    ) -> Iterator[ProductInsight]:
        candidates = [
            prediction
            for prediction in predictions
            if cls.is_prediction_valid(product, prediction.value_tag)  # type: ignore
        ]
        for candidate in candidates:
            if candidate.value_tag:
                # we normalize `value_tag` to the canonical taxonomy value
                # it helps to deal with internationalization
                # (`fr:sans-gluten` and `en:no-gluten` is the same tag)
                value_tag = match_taxonomized_value(
                    candidate.value_tag, TaxonomyType.label.name
                )
                if value_tag is not None:
                    candidate.value_tag = value_tag

        taxonomy = get_taxonomy(InsightType.label.name)
        for candidate in select_deepest_taxonomized_candidates(candidates, taxonomy):
            insight = ProductInsight(**candidate.to_dict())
            if insight.automatic_processing is None:
                insight.automatic_processing = (
                    candidate.value_tag in AUTHORIZED_LABELS_STORE.get()
                )
            yield insight


class CategoryImporter(InsightImporter):
    @staticmethod
    def get_type() -> InsightType:
        return InsightType.category

    @staticmethod
    def get_required_prediction_types() -> set[PredictionType]:
        return {PredictionType.category}

    @classmethod
    def is_conflicting_insight(
        cls, candidate: ProductInsight, reference: ProductInsight
    ) -> bool:
        return candidate.value_tag == reference.value_tag or cls.is_parent_category(
            candidate.value_tag, {reference.value_tag}  # type: ignore
        )

    @classmethod
    def is_parent_category(cls, category: str, to_check_categories: set[str]):
        # Check that the predicted category is not a parent of a
        # current/already predicted category
        return get_taxonomy(InsightType.category.name).is_parent_of_any(
            category, to_check_categories, raises=False
        )

    @classmethod
    def generate_candidates(
        cls,
        product: Product,
        predictions: list[Prediction],
    ) -> Iterator[ProductInsight]:
        candidates = [
            prediction
            for prediction in predictions
            if cls.is_prediction_valid(product, prediction.value_tag)  # type: ignore
        ]
        taxonomy = get_taxonomy(InsightType.category.name)
        yield from (
            ProductInsight(**candidate.to_dict())
            for candidate in select_deepest_taxonomized_candidates(candidates, taxonomy)
        )

    @staticmethod
    def is_prediction_valid(
        product: Product,
        category: str,
    ) -> bool:
        # check whether this is new information or if the predicted category
        # is not a parent of a current/already predicted category
        return not (
            category in product.categories_tags
            or CategoryImporter.is_parent_category(
                category, set(product.categories_tags)
            )
        )

    @classmethod
    def add_optional_fields(cls, insight: ProductInsight, product: Product):
        taxonomy = get_taxonomy(InsightType.category.name)
        if (
            insight.value_tag in taxonomy
            and "agribalyse_food_code" in taxonomy[insight.value_tag].additional_data
        ):
            # This category is linked to an agribalyse category, add it as a
            # campaign tag
            insight.campaign = ["agribalyse-category"]


class ProductWeightImporter(InsightImporter):
    @staticmethod
    def get_type() -> InsightType:
        return InsightType.product_weight

    @staticmethod
    def get_required_prediction_types() -> set[PredictionType]:
        return {PredictionType.product_weight}

    @classmethod
    def is_conflicting_insight(
        cls, candidate: ProductInsight, reference: ProductInsight
    ) -> bool:
        return candidate.value == reference.value

    @staticmethod
    def group_by_subtype(predictions: list[Prediction]) -> dict[str, list[Prediction]]:
        predictions_by_subtype: dict[str, list[Prediction]] = {}

        for prediction in predictions:
            matcher_type = prediction.data["matcher_type"]
            predictions_by_subtype.setdefault(matcher_type, [])
            predictions_by_subtype[matcher_type].append(prediction)

        return predictions_by_subtype

    @classmethod
    def generate_candidates(
        cls,
        product: Product,
        predictions: list[Prediction],
    ) -> Iterator[ProductInsight]:
        if product.quantity is not None or not predictions:
            # Don't generate candidates if the product weight is already
            # specified or if there are no predictions
            return

        # Only generate a single prediction at a time.
        # Predictions are sorted by ascending priority, so the first
        # prediction is assumed to be the best one
        prediction = predictions[0]
        insights_by_subtype = cls.group_by_subtype(predictions)

        insight = ProductInsight(**prediction.to_dict())
        if (
            len(set(x.value for x in insights_by_subtype[insight.data["matcher_type"]]))
            > 1
        ) or insight.data.get("source") == "product_name":
            # Multiple candidates with the same subtype and value, or product
            # weight coming from the product name (less accurate that OCR data)
            # -> don't process automatically
            insight.automatic_processing = False

        yield insight


class ExpirationDateImporter(InsightImporter):
    @staticmethod
    def get_type() -> InsightType:
        return InsightType.expiration_date

    @staticmethod
    def get_required_prediction_types() -> set[PredictionType]:
        return {PredictionType.expiration_date}

    @classmethod
    def is_conflicting_insight(
        cls, candidate: ProductInsight, reference: ProductInsight
    ) -> bool:
        return candidate.value == reference.value

    @classmethod
    def generate_candidates(
        cls,
        product: Product,
        predictions: list[Prediction],
    ) -> Iterator[ProductInsight]:
        if (product and product.expiration_date) or not predictions:
            return

        # expiration date values are formatted according to ISO 8601, so we
        # can sort them in descending order to get the most recent one
        prediction = sorted(
            predictions, key=operator.attrgetter("value"), reverse=True
        )[0]
        if any(prediction.value != other.value for other in predictions):
            prediction.automatic_processing = False
        yield ProductInsight(**prediction.to_dict())


class BrandInsightImporter(InsightImporter):
    @staticmethod
    def get_type() -> InsightType:
        return InsightType.brand

    @staticmethod
    def get_required_prediction_types() -> set[PredictionType]:
        return {PredictionType.brand}

    @classmethod
    def is_conflicting_insight(
        cls, candidate: ProductInsight, reference: ProductInsight
    ) -> bool:
        return candidate.value_tag == reference.value_tag

    @staticmethod
    def is_in_barcode_range(barcode: str, tag: str) -> bool:
        brand_prefix = get_brand_prefix()

        if not in_barcode_range(brand_prefix, tag, barcode):
            logger.info("Barcode %s of brand %s not in barcode range", barcode, tag)
            return False

        return True

    @classmethod
    def generate_candidates(
        cls,
        product: Product,
        predictions: list[Prediction],
    ) -> Iterator[ProductInsight]:
        if product.brands_tags:
            # For now, don't create an insight if a brand has already been provided
            return

        for prediction in predictions:
            if not (
                prediction.predictor == "universal-logo-detector"
                and "username" in prediction.data
            ) and not cls.is_in_barcode_range(
                product.barcode, prediction.value_tag  # type: ignore
            ):
                # Check barcode range for all predictors except logos detected using
                # universal-logo-detector model and annotated manually
                continue
            insight = ProductInsight(**prediction.to_dict())
            if insight.automatic_processing is None:
                # Validation is needed if the weight was extracted from the product name
                # (not as trustworthy as OCR)
                insight.automatic_processing = (
                    prediction.data.get("source") == "product_name"
                )
            yield insight


class StoreInsightImporter(InsightImporter):
    @staticmethod
    def get_type() -> InsightType:
        return InsightType.store

    @staticmethod
    def get_required_prediction_types() -> set[PredictionType]:
        return {PredictionType.store}

    @classmethod
    def is_conflicting_insight(
        cls, candidate: ProductInsight, reference: ProductInsight
    ) -> bool:
        return candidate.value_tag == reference.value_tag

    @classmethod
    def generate_candidates(
        cls,
        product: Product,
        predictions: list[Prediction],
    ) -> Iterator[ProductInsight]:
        for prediction in predictions:
            insight = ProductInsight(**prediction.to_dict())
            insight.automatic_processing = True
            yield insight


def is_valid_product_prediction(
    prediction: Prediction, product: Optional[Product] = None
) -> bool:
    """Return True if the Prediction is valid and can be imported,
    i.e:
       - if the source image (if any) is valid
       - if the product was not deleted

    :param prediction: The Prediction to check
    :param product: The Product fetched from DBProductStore
    :return: Whether the Prediction is valid
    """
    if not product:
        # the product does not exist (deleted)
        logger.info("Prediction of deleted product %s", prediction.barcode)
        return False

    if prediction.source_image and not is_valid_image(
        product.images, prediction.source_image
    ):
        logger.info(
            f"Invalid image for product {product.barcode}: {prediction.source_image}"
        )
        return False

    return True


def create_prediction_model(
    prediction: Prediction,
    server_domain: str,
    timestamp: datetime.datetime,
):
    return {
        "barcode": prediction.barcode,
        "type": prediction.type.name,
        "data": prediction.data,
        "timestamp": timestamp,
        "value_tag": prediction.value_tag,
        "value": prediction.value,
        "source_image": prediction.source_image,
        "automatic_processing": prediction.automatic_processing,
        "server_domain": server_domain,
        "predictor": prediction.predictor,
    }


def import_product_predictions(
    barcode: str,
    product_predictions_iter: Iterable[Prediction],
    server_domain: str,
):
    """Import predictions for a specific product.

    If a prediction already exists in DB (same (barcode, type, server_domain,
    source_image, value, value_tag, predictor, automatic_processing)), it
    won't be imported.

    :param barcode: Barcode of the product. All `product_predictions` must
    have the same barcode.
    :param product_predictions_iter: Iterable of Predictions.
    :param server_domain: The server domain associated with the predictions.
    :return: The number of items imported in DB.
    """
    timestamp = datetime.datetime.utcnow()
    existing_predictions = set(
        PredictionModel.select(
            PredictionModel.type,
            PredictionModel.server_domain,
            PredictionModel.source_image,
            PredictionModel.value_tag,
            PredictionModel.value,
            PredictionModel.predictor,
            PredictionModel.automatic_processing,
        )
        .where(PredictionModel.barcode == barcode)
        .tuples()
    )

    # note: there are some cases
    # when we could decide to replace old predictions of the same key.
    # It's not yet implemented.
    to_import = (
        create_prediction_model(prediction, server_domain, timestamp)
        for prediction in product_predictions_iter
        if (
            prediction.type,
            server_domain,
            prediction.source_image,
            prediction.value_tag,
            prediction.value,
            prediction.predictor,
            prediction.automatic_processing,
        )
        not in existing_predictions
    )
    return batch_insert(PredictionModel, to_import, 50)


IMPORTERS: list[Type] = [
    PackagerCodeInsightImporter,
    LabelInsightImporter,
    CategoryImporter,
    ProductWeightImporter,
    ExpirationDateImporter,
    BrandInsightImporter,
    StoreInsightImporter,
]


def import_insights(
    predictions: Iterable[Prediction],
    server_domain: str,
    product_store: Optional[DBProductStore] = None,
) -> int:
    """Import predictions and generate (and import) insights from these
    predictions.

    :param predictions: an iterable of Predictions to import
    """
    if product_store is None:
        product_store = get_product_store()

    updated_prediction_types_by_barcode = import_predictions(
        predictions, product_store, server_domain
    )
    return import_insights_for_products(
        updated_prediction_types_by_barcode, server_domain, product_store
    )


def import_insights_for_products(
    prediction_types_by_barcode: dict[str, set[PredictionType]],
    server_domain: str,
    product_store: DBProductStore,
) -> int:
    """Re-compute insights for products with new predictions.

    :param prediction_types_by_barcode: a dict that associates each barcode
    with a set of prediction type that were updated
    :param server_domain: The server domain associated with the predictions
    :param product_store: The product store to use

    :return: Number of imported insights
    """
    imported = 0
    for importer in IMPORTERS:
        required_prediction_types = importer.get_required_prediction_types()
        selected_barcodes: list[str] = []
        for barcode, prediction_types in prediction_types_by_barcode.items():
            if prediction_types >= required_prediction_types:
                selected_barcodes.append(barcode)

        if selected_barcodes:
            predictions = [
                Prediction(**p)
                for p in get_product_predictions(
                    selected_barcodes, list(required_prediction_types)
                )
            ]

            for barcode, product_predictions in itertools.groupby(
                sorted(predictions, key=operator.attrgetter("barcode")),
                operator.attrgetter("barcode"),
            ):
                try:
                    with Lock(name=f"robotoff:import:{barcode}", expire=60, timeout=10):
                        imported += importer.import_insights(
                            barcode,
                            list(product_predictions),
                            server_domain,
                            product_store,
                        )
                except LockedResourceException:
                    logger.info(
                        "Couldn't acquire insight import lock, skipping insight import for product %s",
                        barcode,
                    )
                    continue
    return imported


def import_predictions(
    predictions: Iterable[Prediction],
    product_store: DBProductStore,
    server_domain: str,
) -> dict[str, set[PredictionType]]:
    """Check validity and import provided Prediction.

    :param predictions: the Predictions to import
    :param product_store: The product store to use
    :param server_domain: The server domain associated with the predictions
    :return: dict associating each barcode with prediction types that where
    updated in order to re-compute associated insights
    """
    predictions = [
        p
        for p in predictions
        if is_valid_product_prediction(p, product_store[p.barcode])  # type: ignore
    ]

    predictions_imported = 0
    updated_prediction_types_by_barcode: dict[str, set[PredictionType]] = {}
    for barcode, product_predictions_iter in itertools.groupby(
        sorted(predictions, key=operator.attrgetter("barcode")),
        operator.attrgetter("barcode"),
    ):
        product_predictions_group = list(product_predictions_iter)
        predictions_imported += import_product_predictions(
            barcode, product_predictions_group, server_domain
        )
        updated_prediction_types_by_barcode[barcode] = set(
            prediction.type for prediction in product_predictions_group
        )
    logger.info("%s predictions imported", predictions_imported)
    return updated_prediction_types_by_barcode


def refresh_insights(
    barcode: str,
    server_domain: str,
    product_store: Optional[DBProductStore] = None,
) -> int:
    """Refresh all insights for specific product.

    All predictions are fetched, and insights are created/deleted by each
    InsightImporter.

    This is different from `import_insights`, because here, there is no
    prediction creation.  It's just an refresh based on current database
    predictions. It's useful to refresh insights after an Product Opener
    update (some insights may be invalid).

    :param barcode: Barcode of the product.
    :param server_domain: The server domain associated with the predictions.
    :param product_store: The product store to use, defaults to None
    :return: The number of imported insights.
    """
    if product_store is None:
        product_store = get_product_store()

    predictions = [Prediction(**p) for p in get_product_predictions([barcode])]
    prediction_types = set(p.type for p in predictions)

    imported = 0
    for importer in IMPORTERS:
        required_prediction_types = importer.get_required_prediction_types()
        if prediction_types >= required_prediction_types:
            imported += importer.import_insights(
                barcode,
                [p for p in predictions if p.type in required_prediction_types],
                server_domain,
                product_store,
            )

    return imported


def get_product_predictions(
    barcodes: list[str], prediction_types: Optional[list[str]] = None
) -> Iterator[dict]:
    where_clauses = [PredictionModel.barcode.in_(barcodes)]

    if prediction_types is not None:
        where_clauses.append(PredictionModel.type.in_(prediction_types))

    yield from PredictionModel.select().where(*where_clauses).dicts().iterator()



import dataclasses
import gzip
import json
import re
from pathlib import Path
from typing import BinaryIO, Iterable, Optional, Union

from flashtext import KeywordProcessor

from robotoff import settings
from robotoff.prediction.types import Prediction
from robotoff.types import PredictionType
from robotoff.utils import get_logger
from robotoff.utils.cache import CachedStore
from robotoff.utils.text import strip_accents_ascii

from .dataclass import OCRResult


@dataclasses.dataclass(frozen=True)
class City:
    """A city, storing its name, postal code and GPS coordinates."""

    name: str
    """The city name, lower case, no accents, with special characters replaced with
    spaces."""
    postal_code: str
    """The city's postal code. The format depends on the country."""
    coordinates: Optional[tuple[float, float]]
    """The GPS coordinates of the city as a tuple of two floats, or None."""


def load_cities_fr(source: Union[Path, BinaryIO, None] = None) -> set[City]:
    """Load French cities dataset.

    French cities are taken from the La Poste hexasmal dataset:
    https://datanova.legroupe.laposte.fr/explore/dataset/laposte_hexasmal/. The
    source file must be a gzipped-JSON.

    The returned set of cities can contain multiple items with the same name: multiple
    cities can exist with the same name but a different postal code.

    Also, the original dataset may contain multiple items with are not unique with
    regard to the :class:`City` class' attributes: there are additional fields in the
    original dataset which are ignored here. These duplicates are removed.

    Args:
        source (Path or BinaryIO or None, optional, default None): Path to the dataset
            file or open binary stream. If None, the dataset file contained in the
            repo will be used.

    Returns:
        set of City: List of all French cities as `City` objects.

    Raises:
        ValueError: if a postal code is not a valid French postal code (5 digits).
    """
    # JSON file contains a lot of repeated data. An alternative could be to use the
    # CSV file.
    if source is None:
        source = settings.OCR_CITIES_FR_PATH

    # Load JSON content
    with gzip.open(source, "rb") as cities_file:
        json_data = json.load(cities_file)

    # Create City objects
    cities = []
    for item in json_data:
        city_data = item["fields"]
        name = city_data["nom_de_la_commune"].lower()
        postal_code = city_data["code_postal"]
        if not len(postal_code) == 5 or not postal_code.isdigit():
            raise ValueError(
                "{!r}, invalid FR postal code for city {!r}, must be 5-digits "
                "string".format(postal_code, name)
            )
        coords = city_data.get("coordonnees_gps")
        if coords is not None:
            coords = tuple(coords)
        cities.append(City(name, postal_code, coords))

    # Remove duplicates
    return set(cities)


class AddressExtractor:
    """Text processor to extract French addresses based on city name and postal code.

    The main entry point is the `extract_addresses()` method. An `OCRResult` is
    searched for addresses in the following way:

    * The text is prepared by taking it lower case, removing accents, and replacing
      the characters ' and - with " " (space), as city names must follow this format.
    * City names are searched for in the text.
    * For each city name found, its corresponding postal code is searched for in the
      surrounding text, at a maximum distance of `postal_code_search_distance`.
    * If the postal code is found, the match is added to the list of returned
      addresses, along with an extract of the text surrounding the address,
      at a maximum distance of `text_extract_distance`.

    Args:
        cities (iterable of City): Set of cities to search for.
        postal_code_search_distance (int, optional, default 10): Maximum distance
            from a city name to search for a postal code.
        text_extract_distance (int, optional, default 30): Amount of text surrounding a
            detected address to extract for returning.
    """

    def __init__(
        self,
        cities: Iterable[City],
        postal_code_search_distance: int = 10,
        text_extract_distance: int = 30,
    ):
        self.cities = cities
        self.postal_code_search_distance = postal_code_search_distance
        self.text_extract_distance = text_extract_distance

        self.cities_processor = KeywordProcessor()
        for city in self.cities:
            self.cities_processor.add_keyword(city.name, city)

    def extract_addresses(self, content: Union[str, OCRResult]) -> list[Prediction]:
        """Extract addresses from the given OCR result.

        Args:
            content (OCRResult or str): a string or the OCR result to process.

        Returns:
            list of Prediction: List of addresses extracted from the text. Each entry
            is a dictionary with the items: country_code (always "fr"), city_name,
            postal_code and text_extract.
        """
        if isinstance(content, OCRResult):
            text = self.get_text(content)
        else:
            text = content

        text = self.normalize_text(text)
        city_matches = self.find_city_names(text)

        locations = []
        for city, city_start, city_end in city_matches:
            pc_match = self.find_nearby_postal_code(text, city, city_start, city_end)
            if pc_match is None:
                continue

            pc, pc_start, pc_end = pc_match
            address_start = min(city_start, pc_start) - self.text_extract_distance
            address_end = max(city_end, pc_end) + self.text_extract_distance
            text_extract = text[max(0, address_start) : min(len(text), address_end)]

            locations.append(
                Prediction(
                    type=PredictionType.location,
                    data={
                        "country_code": "fr",
                        "city_name": city.name,
                        "postal_code": city.postal_code,
                        "text_extract": text_extract,
                    },
                )
            )

        return locations

    @staticmethod
    def get_text(ocr_result: OCRResult) -> str:
        """Extract text from the OCR result and prepare it.

        Args:
            ocr_result (OCRResult): The OCR result to process.

        Returns:
            str: The text extracted and prepared.
        """
        text = ocr_result.get_full_text()
        if text is None:
            # Using `OCRResult.text_annotations` directly instead of
            # `OCRResult.get_text_annotations()` because the latter contains
            # the text duplicated
            text = ocr_result.text_annotations[0].text
        return text

    @staticmethod
    def normalize_text(text: str) -> str:
        text = text.lower()
        text = strip_accents_ascii(text)
        return text.replace("'", " ").replace("-", " ")

    def find_city_names(self, text: str) -> list[tuple[City, int, int]]:
        """Find all cities from the search set in the text.

        Args:
            text (str): Text to search city names in.

        Returns:
            list of (City, int, int): The list of `City`s which name was found in the
            text, with the start and end indices of their names locations in the
            text. Empty list if none found.
        """
        return self.cities_processor.extract_keywords(text, span_info=True)

    def find_nearby_postal_code(
        self, text: str, city: City, city_start: int, city_end: int
    ) -> Optional[tuple[str, int, int]]:
        """Search for a city's postal code close to its name in the text.

        The postal code is searched at a maximum distance of
        `postal_code_search_distance` from the city name.

        Assumes digit-only postal code, allows non-digit directly next to it. For
        example, for the city "paris" with postal code "75000", "75000 paris" and
        "fr75000 paris" will match.

        Args:
            text (str): The OCR result text.
            city (City): The `City` for which to search the postal code.
            city_start (int): Start index of the city name match in `text`.
            city_end (int): End index of the city name match in `text`.

        Returns:
            (str, int, int) or None: If the `City`'s postal code was found close to
            the city name match, it is returned along with its start and end indices
            in the text. If it was not found, returns None.
        """
        if not city.postal_code.isdigit():
            logger = get_logger(
                "{}.{}".format(self.__module__, self.__class__.__name__)
            )
            logger.error("postal code contains non-digit characters: %s", city)
            return None
        pattern = r"(?:[^0-9]|^)({})(?:[^0-9]|$)".format(city.postal_code)

        sub_start = max(0, city_start - self.postal_code_search_distance)
        sub_end = min(len(text), city_end + self.postal_code_search_distance)
        sub_text = text[sub_start:sub_end]

        match = re.search(pattern, sub_text)
        if match is None:
            return None
        else:
            return match.group(1), sub_start + match.start(1), sub_start + match.end(1)


ADDRESS_EXTRACTOR_STORE = CachedStore(
    lambda: AddressExtractor(load_cities_fr()), expiration_interval=None
)


def find_locations(content: Union[OCRResult, str]) -> list[Prediction]:
    """Find location predictions in the text content.

    See :class:`.AddressExtractor`.

    Args:
        content (OCRResult or str): The content to be searched for locations.

    Returns:
        list of Prediction: See :meth:`.AddressExtractor.extract_addresses`.
    """
    location_extractor: AddressExtractor = ADDRESS_EXTRACTOR_STORE.get()
    return location_extractor.extract_addresses(content)



import collections
import pathlib
from enum import Enum
from typing import Any, Iterable, Optional, Union

import cachetools
import requests

from robotoff import settings
from robotoff.utils import get_logger, http_session, load_json
from robotoff.utils.text import get_tag
from robotoff.utils.types import JSONType

try:
    import networkx
except ImportError:
    networkx = None

logger = get_logger(__name__)


class TaxonomyType(Enum):
    category = 1
    ingredient = 2
    label = 3
    brand = 4
    packaging_shape = 5
    packaging_material = 6
    packaging_recycling = 7


class TaxonomyNode:
    __slots__ = ("id", "names", "parents", "children", "synonyms", "additional_data")

    def __init__(
        self,
        identifier: str,
        names: dict[str, str],
        synonyms: Optional[dict[str, list[str]]],
        additional_data: Optional[dict[str, Any]] = None,
    ):
        self.id: str = identifier
        self.names: dict[str, str] = names
        self.parents: list["TaxonomyNode"] = []
        self.children: list["TaxonomyNode"] = []
        self.additional_data = additional_data or {}

        if synonyms:
            self.synonyms = synonyms
        else:
            self.synonyms = {}

    def is_child_of(self, item: "TaxonomyNode") -> bool:
        """Return True if `item` is a child of `self` in the taxonomy."""
        if not self.parents:
            return False

        if item in self.parents:
            return True

        for parent in self.parents:
            is_parent = parent.is_child_of(item)

            if is_parent:
                return True

        return False

    def is_parent_of(self, item: "TaxonomyNode") -> bool:
        return item.is_child_of(self)

    def is_parent_of_any(self, candidates: Iterable["TaxonomyNode"]) -> bool:
        for candidate in candidates:
            if candidate.is_child_of(self):
                return True

        return False

    def get_parents_hierarchy(self) -> list["TaxonomyNode"]:
        """Return the list of all parent nodes (direct and indirect)."""
        all_parents = []
        seen: set[str] = set()

        if not self.parents:
            return []

        for self_parent in self.parents:
            if self_parent.id not in seen:
                all_parents.append(self_parent)
                seen.add(self_parent.id)

            for parent_parent in self_parent.get_parents_hierarchy():
                if parent_parent.id not in seen:
                    all_parents.append(parent_parent)
                    seen.add(parent_parent.id)

        return all_parents

    def get_localized_name(self, lang: str) -> str:
        if lang in self.names:
            return self.names[lang]

        return self.id

    def get_synonyms(self, lang: str) -> list[str]:
        return self.synonyms.get(lang, [])

    def add_parents(self, parents: Iterable["TaxonomyNode"]):
        for parent in parents:
            if parent not in self.parents:
                self.parents.append(parent)
                parent.children.append(self)

    def to_dict(self) -> JSONType:
        return {"name": self.names, "parents": [p.id for p in self.parents]}

    def __repr__(self):
        return "<TaxonomyNode %s>" % self.id


class Taxonomy:
    def __init__(self):
        self.nodes: dict[str, TaxonomyNode] = {}

    def add(self, key: str, node: TaxonomyNode) -> None:
        self.nodes[key] = node

    def __contains__(self, item: str):
        return item in self.nodes

    def __getitem__(self, item: str):
        return self.nodes.get(item)

    def __len__(self):
        return len(self.nodes)

    def iter_nodes(self) -> Iterable[TaxonomyNode]:
        """Iterate over the nodes of the taxonomy."""
        return iter(self.nodes.values())

    def keys(self):
        return self.nodes.keys()

    def find_deepest_nodes(self, nodes: list[TaxonomyNode]) -> list[TaxonomyNode]:
        """Given a list of nodes, returns the list of nodes where all the parents
        within the list have been removed.

        For example, for a taxonomy, 'fish' -> 'salmon' -> 'smoked-salmon':

        ['fish', 'salmon'] -> ['salmon']
        ['fish', 'smoked-salmon'] -> [smoked-salmon']
        """
        excluded: set[str] = set()

        for node in nodes:
            for second_node in (
                n for n in nodes if n.id not in excluded and n.id != node.id
            ):
                if node.is_child_of(second_node):
                    excluded.add(second_node.id)

        return [node for node in nodes if node.id not in excluded]

    def is_parent_of_any(
        self, item: str, candidates: Iterable[str], raises: bool = True
    ) -> bool:
        """Return True if `item` is parent of any candidate, False otherwise.

        If the item is not in the taxonomy and raises is False, return False.

        :param item: The item to compare
        :param candidates: A list of candidates
        :param raises: if True, raises a ValueError if item is not in the
        taxonomy, defaults to True.
        """
        node: TaxonomyNode = self[item]

        if node is None:
            if raises:
                raise ValueError(f"unknown id in taxonomy: {node}")
            else:
                return False

        to_check_nodes: set[TaxonomyNode] = set()

        for candidate in candidates:
            candidate_node = self[candidate]

            if candidate_node is not None:
                to_check_nodes.add(candidate_node)

        return node.is_parent_of_any(to_check_nodes)

    def get_localized_name(self, key: str, lang: str) -> str:
        if key not in self.nodes:
            return key

        return self.nodes[key].get_localized_name(lang)

    def to_dict(self) -> JSONType:
        export = {}

        for key, node in self.nodes.items():
            export[key] = node.to_dict()

        return export

    @classmethod
    def from_dict(cls, data: JSONType) -> "Taxonomy":
        taxonomy = Taxonomy()

        for key, key_data in data.items():
            if key not in taxonomy:
                node = TaxonomyNode(
                    identifier=key,
                    names=key_data.get("name", {}),
                    synonyms=key_data.get("synonyms", None),
                    additional_data={
                        k: v
                        for k, v in key_data.items()
                        if k not in {"parents", "name", "synonyms", "children"}
                    },
                )
                taxonomy.add(key, node)

        for key, key_data in data.items():
            node = taxonomy[key]
            parents = [taxonomy[ref] for ref in key_data.get("parents", [])]
            node.add_parents(parents)

        return taxonomy

    @classmethod
    def from_json(cls, file_path: Union[str, pathlib.Path]):
        return cls.from_dict(load_json(file_path, compressed=True))  # type: ignore

    def to_graph(self):
        """Generate a networkx.DiGraph from the taxonomy."""
        graph = networkx.DiGraph()
        graph.add_nodes_from((x.id for x in self.iter_nodes()))

        for node in self.iter_nodes():
            for child in node.children:
                graph.add_edge(node.id, child.id)

        return graph


def generate_category_hierarchy(
    taxonomy: Taxonomy, category_to_index: dict[str, int], root: int
):
    categories_hierarchy: dict[int, set] = collections.defaultdict(set)

    for node in taxonomy.iter_nodes():
        category_index = category_to_index[node.id]

        if not node.parents:
            categories_hierarchy[root].add(category_index)

        children_indexes = set(
            [
                category_to_index[c.id]
                for c in node.children
                if c.id in category_to_index
            ]
        )

        categories_hierarchy[category_index] = categories_hierarchy[
            category_index
        ].union(children_indexes)

    categories_hierarchy_list = {}
    for category in categories_hierarchy.keys():
        categories_hierarchy_list[category] = list(categories_hierarchy[category])

    return categories_hierarchy_list


def fetch_taxonomy(
    url: str, fallback_path: Optional[str] = None, offline: bool = False
) -> Taxonomy:
    if offline and fallback_path:
        return Taxonomy.from_json(fallback_path)

    try:
        r = http_session.get(url, timeout=120)  # might take some time
        if r.status_code >= 300:
            raise requests.HTTPError(
                "Taxonomy download at %s returned status code %s", url, r.status_code
            )
        data = r.json()
    except Exception as e:
        logger.exception(f"{type(e)} exception while fetching taxonomy at %s", url)
        if fallback_path:
            return Taxonomy.from_json(fallback_path)
        else:
            raise e

    return Taxonomy.from_dict(data)


@cachetools.cached(cache=cachetools.TTLCache(maxsize=100, ttl=12 * 60 * 60))  # 12h
def get_taxonomy(taxonomy_type: str, offline: bool = False) -> Taxonomy:
    """Returned the requested Taxonomy."""
    if taxonomy_type not in settings.TAXONOMY_URLS:
        raise ValueError(f"unknown taxonomy type: {taxonomy_type}")

    url = settings.TAXONOMY_URLS[taxonomy_type]
    return fetch_taxonomy(
        url,
        fallback_path=str(settings.TAXONOMY_PATHS.get(taxonomy_type, "")) or None,
        offline=offline,
    )


def is_prefixed_value(value: str) -> bool:
    """Return True if the given value has a language prefix (en:, fr:,...),
    False otherwise."""
    return len(value) > 3 and value[2] == ":"


@cachetools.cached(cachetools.TTLCache(maxsize=2, ttl=43200))  # 12h TTL
def get_taxonomy_mapping(taxonomy_type: str) -> dict[str, str]:
    """Return for label type a mapping of prefixed taxonomy values in all
    languages (such as `fr:bio-europeen` or `es:"ecologico-ue`) to their
    canonical value (`en:organic` for the previous example).
    """
    taxonomy = get_taxonomy(taxonomy_type)
    ids: dict[str, str] = {}

    for key in taxonomy.keys():
        if taxonomy_type == TaxonomyType.brand.name:
            unprefixed_key = key
            if is_prefixed_value(key):
                unprefixed_key = key[3:]
            ids[unprefixed_key] = taxonomy[key].names["en"]
        else:
            for lang, name in taxonomy[key].names.items():
                tag = get_tag(name)
                ids[f"{lang}:{tag}"] = key

    return ids


def match_taxonomized_value(value_tag: str, taxonomy_type: str) -> Optional[str]:
    """Return the canonical taxonomized value of a `value_tag` (if any) or
    return None if no match was found or if the type is unsupported.

    Currently it only works for brand and label.
    """
    if taxonomy_type not in (TaxonomyType.brand.name, TaxonomyType.label.name):
        return None

    taxonomy = get_taxonomy(taxonomy_type)

    if value_tag in taxonomy:
        return value_tag

    return get_taxonomy_mapping(taxonomy_type).get(value_tag)


def load_resources():
    """Load and cache resources."""
    logger.info("Loading taxonomy resources...")
    for taxonomy_type in settings.TAXONOMY_URLS.keys():
        get_taxonomy(taxonomy_type)

    for taxonomy_type in (TaxonomyType.brand, TaxonomyType.label):
        get_taxonomy_mapping(taxonomy_type.name)



import functools
from enum import Enum
from typing import Iterable, NamedTuple, Optional, Union

import peewee
from peewee import JOIN, fn

from robotoff import settings
from robotoff.app import events
from robotoff.insights.annotate import (
    ALREADY_ANNOTATED_RESULT,
    SAVED_ANNOTATION_VOTE_RESULT,
    UNKNOWN_INSIGHT_RESULT,
    AnnotationResult,
    InsightAnnotatorFactory,
)
from robotoff.models import (
    AnnotationVote,
    ImageModel,
    ImagePrediction,
    LogoAnnotation,
    Prediction,
    ProductInsight,
    db,
)
from robotoff.off import OFFAuthentication
from robotoff.utils import get_logger

logger = get_logger(__name__)


class SkipVotedType(Enum):
    DEVICE_ID = 1
    USERNAME = 2


class SkipVotedOn(NamedTuple):
    """A helper class to specify whether a voted-on insight should be dropped from
    the get_insights results."""

    by: SkipVotedType
    id: str


def _add_vote_exclusions(
    query: peewee.Query, exclusion: Optional[SkipVotedOn]
) -> peewee.Query:
    if not exclusion:
        return query

    if exclusion.by == SkipVotedType.DEVICE_ID:
        criteria = AnnotationVote.device_id == exclusion.id
    elif exclusion.by == SkipVotedType.USERNAME:
        criteria = AnnotationVote.username == exclusion.id
    else:
        raise ValueError(f"Unknown SkipVoteType: {exclusion.by}")

    return query.join(
        AnnotationVote,
        join_type=peewee.JOIN.LEFT_OUTER,
        on=((AnnotationVote.insight_id == ProductInsight.id) & (criteria)),
    ).where(AnnotationVote.id.is_null())


def get_insights(
    barcode: Optional[str] = None,
    keep_types: Optional[list[str]] = None,
    country: Optional[str] = None,
    brands: Optional[list[str]] = None,
    annotated: Optional[bool] = False,
    annotation: Optional[int] = None,
    order_by: Optional[str] = None,
    value_tag: Optional[str] = None,
    server_domain: Optional[str] = None,
    reserved_barcode: Optional[bool] = None,
    as_dict: bool = False,
    limit: Optional[int] = 25,
    offset: Optional[int] = None,
    count: bool = False,
    avoid_voted_on: Optional[SkipVotedOn] = None,
    group_by_value_tag: Optional[bool] = False,
    automatically_processable: Optional[bool] = None,
    campaign: Optional[str] = None,
    predictor: Optional[str] = None,
) -> Iterable[ProductInsight]:
    if server_domain is None:
        server_domain = settings.OFF_SERVER_DOMAIN

    where_clauses = [ProductInsight.server_domain == server_domain]

    if annotated is not None:
        where_clauses.append(ProductInsight.annotation.is_null(not annotated))

    if annotation is not None:
        where_clauses.append(ProductInsight.annotation == annotation)

    if automatically_processable is not None:
        where_clauses.append(
            ProductInsight.automatic_processing == automatically_processable
        )

    if barcode:
        where_clauses.append(ProductInsight.barcode == barcode)

    if value_tag:
        where_clauses.append(ProductInsight.value_tag == value_tag)

    if keep_types:
        where_clauses.append(ProductInsight.type.in_(keep_types))

    if country is not None:
        where_clauses.append(ProductInsight.countries.contains(country))

    if brands:
        where_clauses.append(ProductInsight.brands.contains_any(brands))

    if reserved_barcode is not None:
        where_clauses.append(ProductInsight.reserved_barcode == reserved_barcode)

    if campaign is not None:
        where_clauses.append(ProductInsight.campaign.contains(campaign))

    if predictor is not None:
        where_clauses.append(ProductInsight.predictor == predictor)

    query = _add_vote_exclusions(ProductInsight.select(), avoid_voted_on)

    if where_clauses:
        query = query.where(*where_clauses)

    if count:
        return query.count()

    if limit is not None:
        query = query.limit(limit)

    if offset is not None and order_by != "random":
        query = query.offset(offset)

    if group_by_value_tag:
        query = query.group_by(ProductInsight.value_tag).order_by(
            fn.COUNT(ProductInsight.id).desc()
        )
        query = query.select(
            ProductInsight.value_tag, fn.Count(ProductInsight.id)
        ).tuples()

    if order_by is not None:
        if order_by == "random":
            # The +1 is here to avoid 0*rand() = 0
            query = query.order_by(
                (peewee.fn.Random() * (ProductInsight.n_votes + 1)).desc()
            )

        elif order_by == "popularity":
            query = query.order_by(ProductInsight.unique_scans_n.desc())

        elif order_by == "n_votes":
            query = query.order_by(ProductInsight.n_votes.desc())

    if as_dict:
        query = query.dicts()

    return query.iterator()


def get_images(
    with_predictions: Optional[bool] = False,
    barcode: Optional[str] = None,
    server_domain: Optional[str] = None,
    offset: Optional[int] = None,
    count: bool = False,
    limit: Optional[int] = 25,
) -> Iterable[ImageModel]:
    if server_domain is None:
        server_domain = settings.OFF_SERVER_DOMAIN

    where_clauses = [ImageModel.server_domain == server_domain]

    if barcode:
        where_clauses.append(ImageModel.barcode == barcode)

    query = ImageModel.select()

    if not with_predictions:
        # return only images without prediction
        query = query.join(ImagePrediction, JOIN.LEFT_OUTER).where(
            ImagePrediction.image.is_null()
        )

    if where_clauses:
        query = query.where(*where_clauses)

    if count:
        return query.count()
    else:
        return query.iterator()


def get_predictions(
    barcode: Optional[str] = None,
    keep_types: Optional[list[str]] = None,
    value_tag: Optional[str] = None,
    server_domain: Optional[str] = None,
    limit: Optional[int] = 25,
    offset: Optional[int] = None,
    count: bool = False,
) -> Iterable[Prediction]:
    if server_domain is None:
        server_domain = settings.OFF_SERVER_DOMAIN

    where_clauses = [Prediction.server_domain == server_domain]

    if barcode:
        where_clauses.append(Prediction.barcode == barcode)

    if value_tag:
        where_clauses.append(Prediction.value_tag == value_tag)

    if keep_types:
        where_clauses.append(Prediction.type.in_(keep_types))

    query = Prediction.select()

    if where_clauses:
        query = query.where(*where_clauses)

    query = query.order_by(Prediction.id.desc())

    if count:
        return query.count()
    else:
        return query.iterator()


def get_image_predictions(
    with_logo: Optional[bool] = False,
    barcode: Optional[str] = None,
    type: Optional[str] = None,
    server_domain: Optional[str] = None,
    offset: Optional[int] = None,
    count: bool = False,
    limit: Optional[int] = 25,
) -> Iterable[ImagePrediction]:

    query = ImagePrediction.select()

    if server_domain is None:
        server_domain = settings.OFF_SERVER_DOMAIN

    query = query.switch(ImagePrediction).join(ImageModel)
    where_clauses = [ImagePrediction.image.server_domain == server_domain]

    if barcode:
        where_clauses.append(ImagePrediction.image.barcode == barcode)

    if type:
        where_clauses.append(ImagePrediction.type == type)

    if not with_logo:
        # return only images without logo
        query = (
            query.switch(
                ImagePrediction
            )  # we need this because we may have joined with ImageModel
            .join(LogoAnnotation, JOIN.LEFT_OUTER)
            .where(LogoAnnotation.image_prediction.is_null())
        )

    if where_clauses:
        query = query.where(*where_clauses)

    query = query.order_by(LogoAnnotation.image_prediction.id.desc())

    if count:
        return query.count()
    else:
        return query.iterator()


def save_annotation(
    insight_id: str,
    annotation: int,
    device_id: str,
    update: bool = True,
    data: Optional[dict] = None,
    auth: Optional[OFFAuthentication] = None,
    trusted_annotator: bool = False,
) -> AnnotationResult:
    """Saves annotation either by using a single response as ground truth or
    by using several responses.

    If annotation == -1 (ignore), we consider the annotation as a vote, so
    that we don't return the insight to the user again.
    If there are >= 2 votes on one of the 2 other possible values (0/1),
    including the vote in process, we set the annotation value of the largest
    voting group. The only exception is when both groups >= 2 votes, in which
    case we mark the insight as invalid (annotation=-1).
    Note that `annotation=-1` has two different meanings here: if it's a vote,
    we consider it as a "ignore", and if it's the insight annotation value we
    consider the insight as invalid, so that it's not available for annotation
    again.

    :param insight_id: The ID of the insight
    :param annotation: The annotation, either -1, 0, or 1
    :param device_id: Unique identifier of the device, see
      `device_id_from_request`
    :param update: If True, perform the update on Product Opener if annotation=1,
      otherwise only save the annotation (default: True)
    :param data: Optional additional data, required for some insight types
    :param auth: User authentication data
    :param trusted_annotator: Defines whether the given annotation comes from
    an authoritative source (e.g. a trusted user), ot whether the annotation
    should be subject to the voting system.
    """
    try:
        insight: Union[ProductInsight, None] = ProductInsight.get_by_id(insight_id)
    except ProductInsight.DoesNotExist:
        insight = None

    if not insight:
        return UNKNOWN_INSIGHT_RESULT

    if insight.annotation is not None:
        return ALREADY_ANNOTATED_RESULT

    # We use AnnotationVote mechanism to save annotation = -1 (ignore) for
    # authenticated users, so that it's not returned again to the user
    if not trusted_annotator or annotation == -1:
        verified = False

        AnnotationVote.create(
            insight_id=insight_id,
            username=auth.get_username() if auth else None,
            value=annotation,
            device_id=device_id,
        )

        with db.atomic() as tx:
            try:
                existing_votes = list(
                    AnnotationVote.select(
                        AnnotationVote.value,
                        peewee.fn.COUNT(AnnotationVote.value).alias("num_votes"),
                    )
                    .where(
                        AnnotationVote.insight_id == insight_id,
                        # We don't consider any ignore (annotation = -1) vote
                        AnnotationVote.value != -1,
                    )
                    .group_by(AnnotationVote.value)
                    .order_by(peewee.SQL("num_votes").desc())
                )
                insight.n_votes = functools.reduce(
                    lambda sum, row: sum + row.num_votes, existing_votes, 0
                )
                insight.save()
            except Exception as e:
                tx.rollback()
                raise e

        if existing_votes:
            # If the top annotation has 3 or more votes, consider applying it to the insight.
            if existing_votes[0].num_votes >= 3:
                annotation = existing_votes[0].value
                verified = True

            # But first check for the following cases:
            #  1) The 1st place annotation has > 2 votes, and the 2nd place annotation has >= 2 votes.
            #  2) 1st place and 2nd place have 2 votes each.
            #
            # In both cases, we consider this an ambiguous result and mark it with 'I don't know'.
            if (
                existing_votes[0].num_votes >= 2
                and len(existing_votes) > 1
                and existing_votes[1].num_votes >= 2
            ):
                # This code credits the last person to contribute a vote with a potentially not their annotation.
                annotation = -1
                verified = True

        if not verified:
            return SAVED_ANNOTATION_VOTE_RESULT

    annotator = InsightAnnotatorFactory.get(insight.type)
    result = annotator.annotate(insight, annotation, update, data=data, auth=auth)
    username = auth.get_username() if auth else "unknown annotator"
    events.event_processor.send_async(
        "question_answered", username, device_id, insight.barcode
    )
    return result


def get_logo_annotation(
    barcode: Optional[str] = None,
    keep_types: Optional[list[str]] = None,
    value_tag: Optional[str] = None,
    server_domain: Optional[str] = None,
    limit: Optional[int] = 25,
    offset: Optional[int] = None,
    count: bool = False,
) -> Iterable[LogoAnnotation]:

    if server_domain is None:
        server_domain = settings.OFF_SERVER_DOMAIN

    query = LogoAnnotation.select().join(ImagePrediction).join(ImageModel)

    where_clauses = [
        LogoAnnotation.image_prediction.image.server_domain == server_domain
    ]

    if barcode:
        where_clauses.append(LogoAnnotation.image_prediction.image.barcode == barcode)

    if value_tag:
        where_clauses.append(LogoAnnotation.annotation_value_tag == value_tag)

    if keep_types:
        where_clauses.append(LogoAnnotation.annotation_type.in_(keep_types))

    if where_clauses:
        query = query.where(*where_clauses)

    if limit is not None:
        query = query.limit(limit)

    if offset is not None:
        query = query.offset(offset)

    query = query.order_by(LogoAnnotation.image_prediction.id.desc())

    if count:
        return query.count()
    else:
        return query.iterator()



import abc
import datetime
from dataclasses import dataclass
from enum import Enum
from typing import Optional

from requests.exceptions import HTTPError, SSLError, Timeout

from robotoff.insights.dataclass import InsightType
from robotoff.insights.normalize import normalize_emb_code
from robotoff.models import ProductInsight, db
from robotoff.off import (
    OFFAuthentication,
    add_brand,
    add_category,
    add_label_tag,
    add_store,
    save_ingredients,
    select_rotate_image,
    update_emb_codes,
    update_expiration_date,
    update_quantity,
)
from robotoff.products import get_image_id, get_product
from robotoff.utils import get_logger

"""
This file allows to annotate products.

To check whether the annotation already exists or not (and save it and send it to the Open Food Facts database), use the following commands:
    from robotoff.insights.annotate import InsightAnnotatorFactor
    annotator = InsightAnnotatorFactory.get(insight_type)
    annotator.annotate(insight: ProductInsight, annotation: int, update: bool = True, data: Optional[dict] = None, auth: Optional[OFFAuthentication] = None, automatic: bool = False)

If you don't want to update the Open Food Facts database but only save the insight annotation (if the update is performed on the client side for example), you can call `annotate()` with `update=False`.
"""


logger = get_logger(__name__)


@dataclass
class AnnotationResult:
    status_code: int
    status: str
    description: Optional[str] = None


class AnnotationStatus(Enum):
    saved = 1
    updated = 2
    error_missing_product = 3
    error_updated_product = 4
    error_already_annotated = 5
    error_unknown_insight = 6
    error_missing_data = 7
    error_invalid_image = 8
    vote_saved = 9
    error_failed_update = 10


SAVED_ANNOTATION_RESULT = AnnotationResult(
    status_code=AnnotationStatus.saved.value,
    status=AnnotationStatus.saved.name,
    description="the annotation was saved",
)
UPDATED_ANNOTATION_RESULT = AnnotationResult(
    status_code=AnnotationStatus.updated.value,
    status=AnnotationStatus.updated.name,
    description="the annotation was saved and sent to OFF",
)
MISSING_PRODUCT_RESULT = AnnotationResult(
    status_code=AnnotationStatus.error_missing_product.value,
    status=AnnotationStatus.error_missing_product.name,
    description="the product could not be found on OFF",
)
ALREADY_ANNOTATED_RESULT = AnnotationResult(
    status_code=AnnotationStatus.error_already_annotated.value,
    status=AnnotationStatus.error_already_annotated.name,
    description="the insight has already been annotated",
)
UNKNOWN_INSIGHT_RESULT = AnnotationResult(
    status_code=AnnotationStatus.error_unknown_insight.value,
    status=AnnotationStatus.error_unknown_insight.name,
    description="unknown insight ID",
)
DATA_REQUIRED_RESULT = AnnotationResult(
    status_code=AnnotationStatus.error_missing_data.value,
    status=AnnotationStatus.error_missing_data.name,
    description="annotation data is required as JSON in `data` field",
)
SAVED_ANNOTATION_VOTE_RESULT = AnnotationResult(
    status_code=AnnotationStatus.vote_saved.value,
    status=AnnotationStatus.vote_saved.name,
    description="the annotation vote was saved",
)
FAILED_UPDATE_RESULT = AnnotationResult(
    status_code=AnnotationStatus.error_failed_update.value,
    status=AnnotationStatus.error_failed_update.name,
    description="Open Food Facts update failed",
)


class InsightAnnotator(metaclass=abc.ABCMeta):
    def annotate(
        self,
        insight: ProductInsight,
        annotation: int,
        update: bool = True,
        data: Optional[dict] = None,
        auth: Optional[OFFAuthentication] = None,
        automatic: bool = False,
    ) -> AnnotationResult:
        with db.atomic() as tx:
            try:
                return self._annotate(
                    insight, annotation, update, data, auth, automatic
                )
            except HTTPError as e:
                if e.response.status_code >= 500:
                    logger.info("HTTPError occurred during OFF update: %s", e)
                    logger.info("Rolling back SQL transaction")
                    tx.rollback()
                    return FAILED_UPDATE_RESULT
                raise e
            except (Timeout, SSLError) as e:
                logger.info(
                    "Error occurred during OFF update: %s, %s", type(e).__name__, e
                )
                logger.info("Rolling back SQL transaction")
                tx.rollback()
                return FAILED_UPDATE_RESULT

    def _annotate(
        self,
        insight: ProductInsight,
        annotation: int,
        update: bool = True,
        data: Optional[dict] = None,
        auth: Optional[OFFAuthentication] = None,
        automatic: bool = False,
    ) -> AnnotationResult:
        if self.is_data_required() and data is None:
            return DATA_REQUIRED_RESULT

        username: Optional[str] = None
        if auth is not None:
            username = auth.get_username()

        insight.username = username
        insight.annotation = annotation
        insight.completed_at = datetime.datetime.utcnow()

        if automatic:
            insight.automatic_processing = True

        if annotation == 1 and update:
            # Save insight before processing the annotation
            insight.save()
            annotation_result = self.process_annotation(
                insight, data=data, auth=auth
            )  # calls the process_annotation function of the class corresponding to the current insight type
        else:
            annotation_result = SAVED_ANNOTATION_RESULT

        if annotation_result.status_code in (
            AnnotationStatus.saved.value,
            AnnotationStatus.updated.value,
            AnnotationStatus.error_invalid_image.value,
            AnnotationStatus.error_missing_product.value,
            AnnotationStatus.error_updated_product.value,
        ):
            insight.annotated_result = annotation_result.status_code
            insight.save()

        return annotation_result

    @abc.abstractmethod
    def process_annotation(
        self,
        insight: ProductInsight,
        data: Optional[dict] = None,
        auth: Optional[OFFAuthentication] = None,
    ) -> AnnotationResult:
        pass

    def is_data_required(self) -> bool:
        return False


class PackagerCodeAnnotator(InsightAnnotator):
    def process_annotation(
        self,
        insight: ProductInsight,
        data: Optional[dict] = None,
        auth: Optional[OFFAuthentication] = None,
    ) -> AnnotationResult:
        emb_code: str = insight.value

        product = get_product(insight.barcode, ["emb_codes"])

        if product is None:
            return MISSING_PRODUCT_RESULT

        emb_codes_str: str = product.get("emb_codes", "")

        emb_codes: list[str] = []
        if emb_codes_str:
            emb_codes = emb_codes_str.split(",")

        if self.already_exists(emb_code, emb_codes):
            return ALREADY_ANNOTATED_RESULT

        emb_codes.append(emb_code)
        update_emb_codes(
            insight.barcode,
            emb_codes,
            server_domain=insight.server_domain,
            insight_id=insight.id,
            auth=auth,
        )
        return UPDATED_ANNOTATION_RESULT

    @staticmethod
    def already_exists(new_emb_code: str, emb_codes: list[str]) -> bool:
        emb_codes = [normalize_emb_code(emb_code) for emb_code in emb_codes]

        normalized_emb_code = normalize_emb_code(new_emb_code)

        if normalized_emb_code in emb_codes:
            return True

        return False


class LabelAnnotator(InsightAnnotator):
    def process_annotation(
        self,
        insight: ProductInsight,
        data: Optional[dict] = None,
        auth: Optional[OFFAuthentication] = None,
    ) -> AnnotationResult:
        product = get_product(insight.barcode, ["labels_tags"])

        if product is None:
            return MISSING_PRODUCT_RESULT

        labels_tags: list[str] = product.get("labels_tags") or []

        if insight.value_tag in labels_tags:
            return ALREADY_ANNOTATED_RESULT

        add_label_tag(
            insight.barcode,
            insight.value_tag,
            insight_id=insight.id,
            server_domain=insight.server_domain,
            auth=auth,
        )
        return UPDATED_ANNOTATION_RESULT


class IngredientSpellcheckAnnotator(InsightAnnotator):
    def process_annotation(
        self,
        insight: ProductInsight,
        data: Optional[dict] = None,
        auth: Optional[OFFAuthentication] = None,
    ) -> AnnotationResult:
        barcode = insight.barcode
        lang = insight.data["lang"]
        field_name = "ingredients_text_{}".format(lang)
        product = get_product(barcode, [field_name])

        if product is None:
            return MISSING_PRODUCT_RESULT

        original_ingredients = insight.data["text"]
        corrected = insight.data["corrected"]
        expected_ingredients = product.get(field_name)

        if expected_ingredients != original_ingredients:
            logger.warning(
                "ingredients have changed since spellcheck insight "
                "creation (product %s)",
                barcode,
            )
            return AnnotationResult(
                status_code=AnnotationStatus.error_updated_product.value,
                status=AnnotationStatus.error_updated_product.name,
                description="the ingredient list has been updated since spellcheck",
            )

        save_ingredients(
            barcode,
            corrected,
            lang=lang,
            insight_id=insight.id,
            auth=auth,
        )
        return UPDATED_ANNOTATION_RESULT


class CategoryAnnotator(InsightAnnotator):
    def process_annotation(
        self,
        insight: ProductInsight,
        data: Optional[dict] = None,
        auth: Optional[OFFAuthentication] = None,
    ) -> AnnotationResult:
        product = get_product(insight.barcode, ["categories_tags"])

        if product is None:
            return MISSING_PRODUCT_RESULT

        categories_tags: list[str] = product.get("categories_tags") or []

        if insight.value_tag in categories_tags:
            return ALREADY_ANNOTATED_RESULT

        category_tag = insight.value_tag
        add_category(
            insight.barcode,
            category_tag,
            insight_id=insight.id,
            server_domain=insight.server_domain,
            auth=auth,
        )
        return UPDATED_ANNOTATION_RESULT


class ProductWeightAnnotator(InsightAnnotator):
    def process_annotation(
        self,
        insight: ProductInsight,
        data: Optional[dict] = None,
        auth: Optional[OFFAuthentication] = None,
    ) -> AnnotationResult:
        product = get_product(insight.barcode, ["quantity"])

        if product is None:
            return MISSING_PRODUCT_RESULT

        quantity: Optional[str] = product.get("quantity") or None

        if quantity is not None:
            return ALREADY_ANNOTATED_RESULT

        update_quantity(
            insight.barcode,
            insight.value,
            insight_id=insight.id,
            server_domain=insight.server_domain,
            auth=auth,
        )
        return UPDATED_ANNOTATION_RESULT


class ExpirationDateAnnotator(InsightAnnotator):
    def process_annotation(
        self,
        insight: ProductInsight,
        data: Optional[dict] = None,
        auth: Optional[OFFAuthentication] = None,
    ) -> AnnotationResult:
        product = get_product(insight.barcode, ["expiration_date"])

        if product is None:
            return MISSING_PRODUCT_RESULT

        current_expiration_date = product.get("expiration_date") or None

        if current_expiration_date:
            return ALREADY_ANNOTATED_RESULT

        update_expiration_date(
            insight.barcode,
            insight.value,
            insight_id=insight.id,
            server_domain=insight.server_domain,
            auth=auth,
        )
        return UPDATED_ANNOTATION_RESULT


class BrandAnnotator(InsightAnnotator):
    def process_annotation(
        self,
        insight: ProductInsight,
        data: Optional[dict] = None,
        auth: Optional[OFFAuthentication] = None,
    ) -> AnnotationResult:
        product = get_product(insight.barcode, ["brands_tags"])

        if product is None:
            return MISSING_PRODUCT_RESULT

        add_brand(
            insight.barcode,
            insight.value,
            insight_id=insight.id,
            server_domain=insight.server_domain,
            auth=auth,
        )

        return UPDATED_ANNOTATION_RESULT


class StoreAnnotator(InsightAnnotator):
    def process_annotation(
        self,
        insight: ProductInsight,
        data: Optional[dict] = None,
        auth: Optional[OFFAuthentication] = None,
    ) -> AnnotationResult:
        product = get_product(insight.barcode, ["stores_tags"])

        if product is None:
            return MISSING_PRODUCT_RESULT

        stores_tags: list[str] = product.get("stores_tags") or []

        if insight.value_tag in stores_tags:
            return ALREADY_ANNOTATED_RESULT

        add_store(
            insight.barcode,
            insight.value,
            insight_id=insight.id,
            server_domain=insight.server_domain,
            auth=auth,
        )
        return UPDATED_ANNOTATION_RESULT


class NutritionImageAnnotator(InsightAnnotator):
    def process_annotation(
        self,
        insight: ProductInsight,
        data: Optional[dict] = None,
        auth: Optional[OFFAuthentication] = None,
    ) -> AnnotationResult:
        product = get_product(insight.barcode, ["code"])

        if product is None:
            return MISSING_PRODUCT_RESULT

        image_id = get_image_id(insight.source_image or "")

        if not image_id:
            return AnnotationResult(
                status_code=AnnotationStatus.error_invalid_image.value,
                status=AnnotationStatus.error_invalid_image.name,
                description="the image is invalid",
            )
        image_key = "nutrition_{}".format(insight.value_tag)
        select_rotate_image(
            barcode=insight.barcode,
            image_id=image_id,
            image_key=image_key,
            rotate=insight.data.get("rotation"),
            server_domain=insight.server_domain,
            auth=auth,
        )
        return UPDATED_ANNOTATION_RESULT


class NutritionTableStructureAnnotator(InsightAnnotator):
    def process_annotation(
        self,
        insight: ProductInsight,
        data: Optional[dict] = None,
        auth: Optional[OFFAuthentication] = None,
    ) -> AnnotationResult:
        insight.data["annotation"] = data
        insight.save()
        return SAVED_ANNOTATION_RESULT

    def is_data_required(self):
        return True


class InsightAnnotatorFactory:
    mapping = {
        InsightType.ingredient_spellcheck.name: IngredientSpellcheckAnnotator(),
        InsightType.packager_code.name: PackagerCodeAnnotator(),
        InsightType.label.name: LabelAnnotator(),
        InsightType.category.name: CategoryAnnotator(),
        InsightType.product_weight.name: ProductWeightAnnotator(),
        InsightType.expiration_date.name: ExpirationDateAnnotator(),
        InsightType.brand.name: BrandAnnotator(),
        InsightType.store.name: StoreAnnotator(),
        InsightType.nutrition_image.name: NutritionImageAnnotator(),
        InsightType.nutrition_table_structure.name: NutritionTableStructureAnnotator(),
    }

    @classmethod
    def get(cls, identifier: str) -> InsightAnnotator:
        if identifier not in cls.mapping:
            raise ValueError("unknown annotator: {}".format(identifier))

        return cls.mapping[identifier]


class InvalidInsight(Exception):
    pass



import datetime
from typing import Iterable, Optional

from PIL import Image

from robotoff.models import ImageModel, ImagePrediction
from robotoff.off import get_source_from_url
from robotoff.prediction import ocr
from robotoff.prediction.object_detection import (
    OBJECT_DETECTION_MODEL_VERSION,
    ObjectDetectionModelRegistry,
)
from robotoff.prediction.ocr.core import get_ocr_result
from robotoff.prediction.types import Prediction
from robotoff.types import ObjectDetectionModel, PredictionType
from robotoff.utils import get_logger, http_session

logger = get_logger(__name__)


DEFAULT_OCR_PREDICTION_TYPES: list[PredictionType] = [
    PredictionType.category,
    PredictionType.label,
    PredictionType.packager_code,
    PredictionType.product_weight,
    PredictionType.image_flag,
    PredictionType.expiration_date,
    PredictionType.brand,
    PredictionType.store,
    PredictionType.nutrient,
    PredictionType.nutrient_mention,
    PredictionType.image_lang,
    PredictionType.image_orientation,
]


PRODUCT_NAME_PREDICTION_TYPES: list[PredictionType] = [
    PredictionType.label,
    PredictionType.product_weight,
    PredictionType.brand,
]


def run_object_detection_model(
    model_name: ObjectDetectionModel,
    image: Image.Image,
    source_image: str,
    threshold: float = 0.1,
) -> Optional[ImagePrediction]:
    """Run a model detection model and save the results in the
    `image_prediction` table.

    An item with the corresponding `source_image` in the `image` table is
    expected to exist. Nothing is done if an image prediction already exists
    in DB for this image and model.

    :param model_name: name of the object detection model to use
    :param image: the input Pillow image
    :param source_image: the source image path (used to fetch the image from
      `image` table)
    :param threshold: the minimum object score above which we keep the object data

    :return: return None if the image does not exist in DB, or the created
      `ImagePrediction` otherwise
    """
    image_instance = ImageModel.get_or_none(source_image=source_image)

    if image_instance is None:
        logger.warning("Missing image in DB for image %s", source_image)
        return None

    existing_image_prediction = ImagePrediction.get_or_none(
        image=image_instance, model_name=model_name.value
    )
    if existing_image_prediction is not None:
        logger.info(
            f"Object detection results for {model_name} already exist for "
            f"image {source_image}: ID {existing_image_prediction.id}"
        )
        return None

    timestamp = datetime.datetime.utcnow()
    results = ObjectDetectionModelRegistry.get(model_name.value).detect_from_image(
        image, output_image=False
    )
    data = results.to_json(threshold=threshold)
    max_confidence = max([item["score"] for item in data], default=None)
    return ImagePrediction.create(
        image=image_instance,
        type="object_detection",
        model_name=model_name.value,
        model_version=OBJECT_DETECTION_MODEL_VERSION[model_name],
        data={"objects": data},
        timestamp=timestamp,
        max_confidence=max_confidence,
    )


def get_predictions_from_product_name(
    barcode: str, product_name: str
) -> list[Prediction]:
    predictions_all = []
    for prediction_type in PRODUCT_NAME_PREDICTION_TYPES:
        predictions = ocr.extract_predictions(
            product_name, prediction_type, barcode=barcode
        )
        for prediction in predictions:
            prediction.data["source"] = "product_name"
            # Predictions from product name are not as trustworthy as
            # predictions from OCR, so disable automatic processing
            prediction.automatic_processing = False
        predictions_all += predictions

    return predictions_all


def extract_ocr_predictions(
    barcode: str, ocr_url: str, prediction_types: Iterable[PredictionType]
) -> list[Prediction]:
    logger.info("Generating OCR predictions from OCR %s", ocr_url)

    predictions_all: list[Prediction] = []
    source_image = get_source_from_url(ocr_url)
    ocr_result = get_ocr_result(ocr_url, http_session, error_raise=False)

    if ocr_result is None:
        return predictions_all

    for prediction_type in prediction_types:
        predictions_all += ocr.extract_predictions(
            ocr_result, prediction_type, barcode=barcode, source_image=source_image
        )

    return predictions_all



import contextlib
import gzip
import sys
from pathlib import Path
from typing import Iterable, Optional

import _io
import click
import dacite
import orjson
import tqdm

from robotoff.off import get_barcode_from_path
from robotoff.prediction.ocr import OCRResult, extract_predictions
from robotoff.prediction.ocr.core import ocr_content_iter
from robotoff.prediction.types import Prediction
from robotoff.types import PredictionType
from robotoff.utils import get_logger, gzip_jsonl_iter, jsonl_iter

logger = get_logger(__name__)


def run_from_ocr_archive(
    input_path: Path,
    prediction_type: PredictionType,
    output: Optional[Path] = None,
):
    predictions = tqdm.tqdm(
        generate_from_ocr_archive(input_path, prediction_type), desc="OCR"
    )
    output_f: _io._TextIOBase
    need_decoding = False

    if output is not None:
        if output.suffix == ".gz":
            output_f = gzip.open(output, "wb")
        else:
            output_f = output.open("wb")
    else:
        output_f = sys.stdout
        need_decoding = True

    with contextlib.closing(output_f):
        for prediction in predictions:
            raw_data = orjson.dumps(prediction.to_dict()) + b"\n"
            data = raw_data.decode("utf-8") if need_decoding else raw_data
            output_f.write(data)


def generate_from_ocr_archive(
    input_path: Path,
    prediction_type: PredictionType,
) -> Iterable[Prediction]:
    json_iter = (
        gzip_jsonl_iter(input_path)
        if input_path.suffix == ".gz"
        else jsonl_iter(input_path)
    )
    for source_image, ocr_json in ocr_content_iter(json_iter):
        if source_image is None:
            continue

        barcode: Optional[str] = get_barcode_from_path(source_image)

        if barcode is None:
            click.echo(
                f"cannot extract barcode from source {source_image}",
                err=True,
            )
            continue

        ocr_result: Optional[OCRResult] = OCRResult.from_json(ocr_json)

        if ocr_result is None:
            continue

        yield from extract_predictions(
            ocr_result, prediction_type, barcode=barcode, source_image=source_image
        )


def insights_iter(file_path: Path) -> Iterable[Prediction]:
    for prediction in jsonl_iter(file_path):
        yield dacite.from_dict(
            data_class=Prediction,
            data=prediction,
            config=dacite.Config(cast=[PredictionType]),
        )



"""
Helper to download weights files for different Robotoff models.
"""
import shutil
from pathlib import Path

import requests

from robotoff.utils import get_logger

logger = get_logger(__name__)


def download_file(url: str, destination: Path, force: bool = False) -> None:
    """
    Efficiently download a file from an URL.

    Taken from https://stackoverflow.com/a/39217788.
    """
    # Check if file exists
    if destination.is_file():
        logger.info(f"Destination file already exists: {destination}")

        if not force:
            logger.info(f"Skip download from {url}")
            return
        else:
            logger.info("--force used")

    # Create directory if needed
    file_dir = destination.parent
    if not file_dir.is_dir():
        logger.info(f"Create directory: {file_dir}")
        file_dir.mkdir(exist_ok=True, parents=True)

    # Download file
    logger.info(f"Start download from {url} to {destination}.")
    with requests.get(url, stream=True) as r:
        with destination.open("wb") as f:
            shutil.copyfileobj(r.raw, f)
    logger.info("Download complete.")



from robotoff.settings import FR_TOKENS_PATH, INGREDIENT_TOKENS_PATH
from robotoff.utils import cache, text_file_iter


def get_fr_known_tokens() -> set[str]:
    tokens = set(text_file_iter(INGREDIENT_TOKENS_PATH, comment=False))
    tokens = tokens.union(set(text_file_iter(FR_TOKENS_PATH, comment=False)))
    return tokens


FR_KNOWN_TOKENS_CACHE = cache.CachedStore(get_fr_known_tokens)



from typing import Callable, Iterable, Optional

from flashtext import KeywordProcessor


def generate_keyword_processor(
    items: Iterable[str], keep_func: Optional[Callable] = None
) -> KeywordProcessor:
    processor = KeywordProcessor()

    for item in items:
        splitted = item.split("||")

        if len(splitted) == 2:
            key, name = splitted
            pattern = name
        elif len(splitted) == 3:
            key, name, pattern = splitted
        else:
            raise ValueError("invalid syntax: '{}'".format(item))

        if keep_func is not None and not keep_func(key, name):
            continue

        processor.add_keyword(pattern, clean_name=(key, name))

    return processor



from pymongo import MongoClient

from robotoff import settings
from robotoff.utils.cache import CachedStore


def get_mongo_client() -> MongoClient:
    return MongoClient(settings.MONGO_URI, serverSelectionTimeoutMS=5_000)


MONGO_CLIENT_CACHE = CachedStore(get_mongo_client, expiration_interval=None)



import requests
from healthcheck import HealthCheck
from influxdb_client import InfluxDBClient
from playhouse.postgres_ext import PostgresqlExtDatabase
from pymongo import MongoClient
from pymongo.errors import ServerSelectionTimeoutError
from redis import Redis

from robotoff import settings
from robotoff.utils import get_logger

health = HealthCheck()

logger = get_logger(__name__)


def test_connect_mongodb():
    logger.debug("health: testing mongodb connection to %s", settings.MONGO_URI)
    client = MongoClient(settings.MONGO_URI, serverSelectionTimeoutMS=5_000)
    try:
        client.server_info()
    except ServerSelectionTimeoutError:
        return False, "MongoDB DB connection failed!"
    return True, "MongoDB DB connection succeeded!"


def test_connect_redis():
    logger.debug("health: testing Redis connection to %s", settings.REDIS_HOST)
    client = Redis(host=settings.REDIS_HOST)
    client.ping()
    return True, "Redis DB connection success!"


def test_connect_postgres():
    logger.debug("health: testing postgres connection to %s", settings.POSTGRES_HOST)
    client = PostgresqlExtDatabase(
        settings.POSTGRES_DB,
        user=settings.POSTGRES_USER,
        password=settings.POSTGRES_PASSWORD,
        host=settings.POSTGRES_HOST,
        port=5432,
    )
    client.connect()
    return True, "Postgres db connection success !"


def test_connect_influxdb():
    if not settings.INFLUXDB_HOST:
        return True, "skipped InfluxDB db connection test (INFLUXDB_HOST is empty)!"
    logger.debug("health: testing InfluxDB connection to %s", settings.INFLUXDB_HOST)
    client = InfluxDBClient(
        url=f"http://{settings.INFLUXDB_HOST}:{settings.INFLUXDB_PORT}",
        token=settings.INFLUXDB_AUTH_TOKEN,
        org=settings.INFLUXDB_ORG,
    )
    if not client.ping():
        return False, "InfluxDB connection failed!"
    return True, "InfluxDB db connection success!"


def test_connect_robotoff_api():
    logger.debug("health: testing robotoff API status")
    resp = requests.get(f"{settings.BaseURLProvider().robotoff().get()}/api/v1/status")
    return resp.json()["status"] == "running", "Robotoff API connection success !"


def test_connect_ann():
    logger.debug("health: testing robotoff ann status")
    resp = requests.get(
        f"{settings.BaseURLProvider().robotoff().get()}/api/v1/ann/?count=1"
    )
    return "count" in resp.json(), "ANN API connection success !"


health.add_check(test_connect_mongodb)
health.add_check(test_connect_postgres)
health.add_check(test_connect_influxdb)
health.add_check(test_connect_redis)
health.add_check(test_connect_robotoff_api)
health.add_check(test_connect_ann)



from typing import Union

from robotoff.prediction.types import Prediction
from robotoff.types import PredictionType

from .dataclass import ImageOrientation, OCRResult


def get_rotation_angle_from_orientation(image_orientation: ImageOrientation) -> int:
    if image_orientation == ImageOrientation.up:
        return 0
    elif image_orientation == ImageOrientation.unknown:
        return 0
    elif image_orientation == ImageOrientation.left:
        return 90
    elif image_orientation == ImageOrientation.right:
        return 270
    elif image_orientation == ImageOrientation.down:
        return 180
    else:
        raise ValueError("unknown image orientation: {}".format(image_orientation))


def find_image_orientation(ocr_result: Union[OCRResult, str]) -> list[Prediction]:
    if isinstance(ocr_result, str):
        return []

    orientation_result = ocr_result.get_orientation()

    if orientation_result is None:
        return []

    prediction = orientation_result.to_json()
    prediction["rotation"] = get_rotation_angle_from_orientation(
        orientation_result.orientation
    )
    return [Prediction(type=PredictionType.image_orientation, data=prediction)]



from robotoff.utils.text import strip_accents_ascii


def normalize_emb_code(emb_code: str):
    """replacing space, hyphen, and dot
    so that "fr 40.261.001 ce" changes to "fr40261001ce'"
    """

    emb_code = (
        emb_code.strip().lower().replace(" ", "").replace("-", "").replace(".", "")
    )

    emb_code = strip_accents_ascii(emb_code)

    """if the code ends with "ce" replace it with "ec"
    here "fr40261001ce" becomes "fr40261001ec"
    """

    if emb_code.endswith("ce"):
        emb_code = emb_code[:-2] + "ec"

    return emb_code
